namespace BonkTurnip
{
	void AddAir(CBaseEntity@ pActivator, CBaseEntity@ pCaller, USE_TYPE useType, float flValue)
	{
		pActivator.pev.air_finished = g_Engine.time + 30;
	}


	// Below code added by Erty to implement a scrap collected
	// progress message displayed in the HUD.

    /**
     * Map activation handler.
     * @return void
     */
    void MapActivate()
    {
        @g_pInstance = MapPart();

        if (!g_pInstance.Initialise()) {
            g_Game.AlertMessage(at_console, "[Ragemap 2024] Bonk's part encountered errors initialising.\n");

            @g_pInstance = null;
            return;
        }

        g_Game.AlertMessage(at_console, "[Ragemap 2024] Bonk's part ready.\n");
    }


    /**
     * Ragemap 2024: Bonk's Part.
     */
    final class MapPart
    {
        /*
         * -------------------------------------------------------------------------
         * Variables
         * -------------------------------------------------------------------------
         */

        /** @var EHandle m_hScrapCounter It counts scrap */
        EHandle m_hScrapCounter;
		/** @var int m_iScrapLimit How much scrap we need to collect */
        int m_iScrapLimit;
		/** @var bool m_fFinished True if all scrap has been collected */
        bool m_fFinished;
		/** @var HUDTextParams msgParams HUD text params object */
		HUDTextParams msgParams;

        /**
         * Constructor.
         */
        MapPart()
        {
			m_fFinished = false;

			msgParams.x           = -1;
            msgParams.y           = 0.15;
            msgParams.effect      = 0;
            msgParams.r1          = 100;
            msgParams.g1          = 100;
            msgParams.b1          = 100;
            msgParams.a1          = 0;
            msgParams.r2          = 240;
            msgParams.g2          = 100;
            msgParams.b2          = 0;
            msgParams.a2          = 0;
            msgParams.holdTime    = 0.5;
			msgParams.fadeinTime  = 0.0;
			msgParams.fadeoutTime = 0.5;
            msgParams.channel     = 2;
            // ... initialise variables (just above) with defaults values here
        }

        /**
         * Initialise.
         * @return bool Success
         */
        bool Initialise()
        {
            uint uiErrors = 0;
            CBaseEntity@ pEntity;

            @pEntity = null;
            if ((@pEntity = g_EntityFuncs.FindEntityByTargetname(pEntity, "bonk_scrap_counter_1")) !is null) {
                m_hScrapCounter = EHandle(pEntity);
				m_iScrapLimit = int(pEntity.pev.health);
            } else {
                g_Game.AlertMessage(at_error, "Ragemap2024::Bonk::MapPart->Initialise(): Failed to find scrap game_counter.\n");
                ++uiErrors;
            }

            // ... perform some basic pre-flight checks here if needed
            return (0 == uiErrors);
        }

        // ... public functions you would call from outside the class here

		void PrintScrapProgress()
		{
			if (!m_hScrapCounter.IsValid())
            {
                g_Game.AlertMessage(at_error, "Ragemap2024::Bonk->PrintScrapProgress(): Scrap counter not found\n");
                return;
            }

			CBaseEntity@ pScrapCounter = m_hScrapCounter.GetEntity();

			string scrapCounterMessage;

			if (m_fFinished) {
				scrapCounterMessage = "";
			} else {
				snprintf(scrapCounterMessage, "Scrap remaining: %1", m_iScrapLimit - pScrapCounter.pev.frags);
			}
			
			msgParams.channel = 2;
			g_PlayerFuncs.HudMessageAll(msgParams, scrapCounterMessage);
		}

		void SetFinished()
		{
			m_fFinished = true;
		}
    }

    MapPart@ g_pInstance;


    /**
     * Map hook: PrintScrapProgress
     * @param  CBaseEntity@|null pActivator Activator entity
     * @return void
     */
    void PrintScrapProgress(CBaseEntity@ pTriggerScript)
    {
        if (g_pInstance is null) {
            return;
        }
        g_pInstance.PrintScrapProgress();
    }

    /**
     * Map hook: Set finished to true
     * @param  CBaseEntity@|null pActivator Activator entity
     * @param  CBaseEntity@|null pCaller    Caller entity
     * @param  USE_TYPE          useType    Use type, or unspecified to assume `USE_TOGGLE`
     * @param  float             flValue    Use value, or unspecified to assume `0.0f`
     * @return void
     */
    void SetFinished(CBaseEntity@ pActivator, CBaseEntity@ pCaller, USE_TYPE useType, float flValue)
    {
        if (g_pInstance is null) {
            return;
        }
        g_pInstance.SetFinished();
    }

}
