#include "erty_lung"

namespace Ragemap2024Erty
{
    array<string> peels = {
        "ragemap2024/erty/bananapeel0.ogg",
        "ragemap2024/erty/bananapeel1.ogg",
        "ragemap2024/erty/bananapeel2.ogg",
        "ragemap2024/erty/bananapeel3.ogg",
        "ragemap2024/erty/bananapeel4.ogg",
        "ragemap2024/erty/bananapeel5.ogg",
        "ragemap2024/erty/bananapeel6.ogg",
        "ragemap2024/erty/bananapeel7.ogg"
    };

    /**
     * Map initialisation handler.
     * @return void
     */
    void MapInit()
    {
        // g_Module.ScriptInfo.SetAuthor("Erty");
        g_CustomEntityFuncs.RegisterCustomEntity("func_erty_lung", "func_erty_lung");
        g_SoundSystem.PrecacheSound("buttons/bell1.wav");

        g_SoundSystem.PrecacheSound("ragemap2024/erty/bananapeel0.ogg");
        g_SoundSystem.PrecacheSound("ragemap2024/erty/bananapeel1.ogg");
        g_SoundSystem.PrecacheSound("ragemap2024/erty/bananapeel2.ogg");
        g_SoundSystem.PrecacheSound("ragemap2024/erty/bananapeel3.ogg");
        g_SoundSystem.PrecacheSound("ragemap2024/erty/bananapeel4.ogg");
        g_SoundSystem.PrecacheSound("ragemap2024/erty/bananapeel5.ogg");
        g_SoundSystem.PrecacheSound("ragemap2024/erty/bananapeel6.ogg");
        g_SoundSystem.PrecacheSound("ragemap2024/erty/bananapeel7.ogg");
    }

    /**
     * Map activation handler.
     * @return void
     */
    void MapActivate()
    {
        @g_pInstance = MapPart();

        if (!g_pInstance.Initialise()) {
            g_Game.AlertMessage(at_console, "[Ragemap 2024] Erty's part encountered errors initialising.\n");

            @g_pInstance = null;
            return;
        }

        g_Game.AlertMessage(at_console, "[Ragemap 2024] Erty's part ready.\n");
    }


    /**
     * Ragemap 2024: Erty's Part.
     */
    final class MapPart
    {
        /*
         * -------------------------------------------------------------------------
         * Variables
         * -------------------------------------------------------------------------
         */

        /** @var EHandle m_hLung Used to absorb oxygen from the air and expell carbondioxide */
        EHandle m_hLung;
        /** @var EHandle m_hLungHome The Lung's home */
        EHandle m_hLungHome;
        /** @var EHandle m_hCompass Lung compass */
        EHandle m_hCompass;
        /** @var EHandle m_hTeleZone Meet Your Master */
        EHandle m_hTeleZone;
        /** @var EHandle m_hDepthIndicator Better not go too deep */
        EHandle m_hDepthIndicator;
        /** @var bool m_fLungForward Whether Lung is moving */
        bool m_fLungForward;
        /** @var array<EHandle> m_rnhNoParkings Places the Lung can't park */
        array<EHandle> m_rnhNoParkings;
        /** @var array<EHandle> m_rnhArtifacts Ooooh nifty thingies */
        array<EHandle> m_rnhArtifacts;
        /** @var float m_fLungInitialDepth The Lung was here */
        float m_fLungInitialDepth;
        /** @var float m_fIndicatorInitialDepth The indicator was here */
        float m_fIndicatorInitialDepth;
        /** @var float m_fLungSinkSpeed The Lung is negatively bouyant */
        float m_fLungSinkSpeed;
        /** @var float m_fCameraRange Optimal range for spicy pictures */
        float m_fCameraRange;

        /**
         * Constructor.
         */
        MapPart()
        {
            m_fLungSinkSpeed = -0.125;
            m_fCameraRange = 20.0;
            m_fLungForward = false;
            // ... initialise variables (just above) with defaults values here
        }

        /**
         * Initialise.
         * @return bool Success
         */
        bool Initialise()
        {
            uint uiErrors = 0;
            CBaseEntity@ pEntity;

            @pEntity = null;
            if ((@pEntity = g_EntityFuncs.FindEntityByTargetname(pEntity, "erty_lung")) !is null) {
                m_hLung = EHandle(pEntity);
                m_fLungInitialDepth = pEntity.pev.origin.z;
            } else {
                g_Game.AlertMessage(at_error, "Ragemap2024::Erty::MapPart->Initialise(): Lung not found.\n");
                ++uiErrors;
            }

            @pEntity = null;
            if ((@pEntity = g_EntityFuncs.FindEntityByTargetname(pEntity, "erty_lung_home")) !is null) {
                m_hLungHome = EHandle(pEntity);
            } else {
                g_Game.AlertMessage(at_error, "Ragemap2024::Erty::MapPart->Initialise(): Lung's home not found.\n");
                ++uiErrors;
            }

            @pEntity = null;
            if ((@pEntity = g_EntityFuncs.FindEntityByTargetname(pEntity, "erty_compass")) !is null) {
                m_hCompass = EHandle(pEntity);
            } else {
                g_Game.AlertMessage(at_error, "Ragemap2024::Erty::MapPart->Initialise(): Compass not found.\n");
                ++uiErrors;
            }

            @pEntity = null;
            if ((@pEntity = g_EntityFuncs.FindEntityByTargetname(pEntity, "erty_telezone")) !is null) {
                m_hTeleZone = EHandle(pEntity);
            } else {
                g_Game.AlertMessage(at_error, "Ragemap2024::Erty::MapPart->Initialise(): TeleZone not found.\n");
                ++uiErrors;
            }

            @pEntity = null;
            if ((@pEntity = g_EntityFuncs.FindEntityByTargetname(pEntity, "erty_lungdepthindicator")) !is null) {
                m_hDepthIndicator = EHandle(pEntity);
                m_fIndicatorInitialDepth = pEntity.pev.origin.z;
            } else {
                g_Game.AlertMessage(at_error, "Ragemap2024::Erty::MapPart->Initialise(): Depth indicator not found.\n");
                ++uiErrors;
            }

            @pEntity = null;
            while ((@pEntity = g_EntityFuncs.FindEntityByTargetname(pEntity, "erty_noparking")) !is null) {
                m_rnhNoParkings.insertLast(EHandle(pEntity));
            }
            if (m_rnhNoParkings.length() < 1) {
                g_Game.AlertMessage(at_error, "Ragemap2024::Erty::MapPart->Initialise(): We can park anywhere, and that ain't right\n");
                ++uiErrors;
            }

            @pEntity = null;
            while ((@pEntity = g_EntityFuncs.FindEntityByTargetname(pEntity, "erty_artifact")) !is null) {
                m_rnhArtifacts.insertLast(EHandle(pEntity));
            }
            if (m_rnhArtifacts.length() < 1) {
                g_Game.AlertMessage(at_error, "Ragemap2024::Erty::MapPart->Initialise(): No artifacts found\n");
                ++uiErrors;
            }

            // ... perform some basic pre-flight checks here if needed
            return (0 == uiErrors);
        }

        // ... public functions you would call from outside the class here

        void UpdateLungCoords()
        {
            if (!m_hLung.IsValid())
            {
                g_Game.AlertMessage(at_error, "Ragemap2024::Erty::MapPart->UpdateLungCoords(): Lung not found\n");
                return;
            }
            if (!m_hLungHome.IsValid())
            {
                g_Game.AlertMessage(at_error, "Ragemap2024::Erty::MapPart->UpdateLungCoords(): Lung's home not found\n");
                return;
            }
            if (!m_hTeleZone.IsValid())
            {
                g_Game.AlertMessage(at_error, "Ragemap2024::Erty::MapPart->UpdateLungCoords(): TeleMaster not found\n");
                return;
            }
            if (!m_hCompass.IsValid())
            {
                g_Game.AlertMessage(at_error, "Ragemap2024::Erty::MapPart->UpdateLungCoords(): Compass not found\n");
                return;
            }

            CBaseEntity@ pLung = m_hLung.GetEntity();
            CustomKeyvalues@ pCustomKVs = pLung.GetCustomKeyvalues();

            CBaseEntity@ pLungHome = m_hLungHome.GetEntity();

            float xCoord = pLung.pev.origin.x;
            float yCoord = pLung.pev.origin.y;
            float left =   pLungHome.pev.absmin.x;
            float right =  pLungHome.pev.absmax.x;
            float bottom = pLungHome.pev.absmin.y;
            float top =    pLungHome.pev.absmax.y;

            pCustomKVs.SetKeyvalue("$f_x", 100.0 * (xCoord - left) / (right - left));
            pCustomKVs.SetKeyvalue("$f_y", 100.0 * (yCoord - bottom) / (top - bottom));

            if (pLung.pev.health <= 50) {
                SinkOrFloat(m_fLungSinkSpeed * 1.2);
            } else {
                SinkOrFloat(m_fLungSinkSpeed);
            }

            CBaseEntity@ pCompass = m_hCompass.GetEntity();

            pLung.pev.angles.y = pCompass.pev.angles.z;

            CBaseEntity@ pTeleZone = m_hTeleZone.GetEntity();
            if (pTeleZone.Intersects(pLung))
            {
                CustomKeyvalues@ pTeleCustomKVs = pTeleZone.GetCustomKeyvalues();

                CustomKeyvalue kvUnlocked(pTeleCustomKVs.GetKeyvalue("$i_unlocked"));
                int unlocked = kvUnlocked.GetInteger();

                if (unlocked == 3)
                {
                    CustomKeyvalue kvTarget(pTeleCustomKVs.GetKeyvalue("$s_target"));
                    string target = kvTarget.GetString();

                    CBaseEntity@ pTarget = null;
                    if ((@pTarget = g_EntityFuncs.FindEntityByTargetname(pTarget, target)) is null)
                    {
                        g_Game.AlertMessage(at_error, "Ragemap2024::Erty::MapPart->UpdateLungCoords: Tele target not found\n");
                        return;
                    }

                    g_EntityFuncs.FireTargets("erty_teleseq", null, null, USE_TOGGLE);
                    g_PlayerFuncs.ScreenShakeAll(pLung.pev.origin, 13, 32, 1.7);

                    pLung.pev.origin = pTarget.pev.origin;
                }
            }
        }

        void SinkOrFloat(float deltaDepth = m_fLungSinkSpeed)
        {
            if (!m_hLung.IsValid())
            {
                g_Game.AlertMessage(at_error, "Ragemap2024::Erty::MapPart->SinkOrFloat(): Lung not found\n");
                return;
            }
            if (!m_hDepthIndicator.IsValid())
            {
                g_Game.AlertMessage(at_error, "Ragemap2024::Erty::MapPart->SinkOrFloat(): Depth indicator not found\n");
                return;
            }

            CBaseEntity@ pLung = m_hLung.GetEntity();
            CustomKeyvalues@ pCustomKVs = pLung.GetCustomKeyvalues();
            CustomKeyvalue kvDepthRange(pCustomKVs.GetKeyvalue("$f_depthRange"));
            CustomKeyvalue kvCrushDepth(pCustomKVs.GetKeyvalue("$f_crushDepth"));
            float depthRange = kvDepthRange.GetFloat();
            float crushDepth = kvCrushDepth.GetFloat();

            float minDepth = m_fLungInitialDepth - depthRange;
            float maxDepth = m_fLungInitialDepth + depthRange;

            pLung.pev.origin.z += deltaDepth;

            if (pLung.pev.origin.z > maxDepth)
            {
                pLung.pev.origin.z = maxDepth;
            }
            else if (pLung.pev.origin.z < minDepth)
            {
                pLung.pev.origin.z = minDepth;
            }

            float relativeDepth = (pLung.pev.origin.z - minDepth) / (maxDepth - minDepth);

            CBaseEntity@ pDepthIndicator = m_hDepthIndicator.GetEntity();
            CustomKeyvalues@ pDepthIndicatorCustomKVs = pDepthIndicator.GetCustomKeyvalues();
            CustomKeyvalue kvIndicatorDepthRange(pDepthIndicatorCustomKVs.GetKeyvalue("$f_range"));
            float indicatorDepthRange = kvIndicatorDepthRange.GetFloat();
            
            float minIndicatorDepth = m_fIndicatorInitialDepth - indicatorDepthRange;
            float maxIndicatorDepth = m_fIndicatorInitialDepth + indicatorDepthRange;

            pDepthIndicator.pev.origin.z = minIndicatorDepth
                + (relativeDepth * (maxIndicatorDepth - minIndicatorDepth));

            return;
        }

        void TakePhoto()
        {
            if (!m_hLung.IsValid())
            {
                g_Game.AlertMessage(at_error, "Ragemap2024::Erty::MapPart->TakePhoto(): Lung not found\n");
                return;
            }
            if (!m_hTeleZone.IsValid())
            {
                g_Game.AlertMessage(at_error, "Ragemap2024::Erty::MapPart->TakePhoto(): TeleMaster not found\n");
                return;
            }

            CBaseEntity@ pLung = m_hLung.GetEntity();
            float lungDepth = pLung.pev.origin.z;

            // Check that we're within optimal range for pictures
            if ((lungDepth < m_fLungInitialDepth - m_fCameraRange) || (lungDepth > m_fLungInitialDepth + m_fCameraRange))
            {
                return;
            }

            for (uint i = 0; i < m_rnhArtifacts.length(); i++)
            {
                if (m_rnhArtifacts[i].IsValid())
                {
                    CBaseEntity@ pArtifact = m_rnhArtifacts[i].GetEntity();
                    if (pArtifact.Intersects(pLung))
                    {
                        CustomKeyvalues@ pCustomKVs = pArtifact.GetCustomKeyvalues();

                        CustomKeyvalue kvDiscovered(pCustomKVs.GetKeyvalue("$i_discovered"));
                        int discovered = kvDiscovered.GetInteger();

                        if (discovered == 1) { return; }

                        CustomKeyvalue kvTarget(pCustomKVs.GetKeyvalue("$s_target"));
                        string artifactTarget = kvTarget.GetString();

                        CBaseEntity@ pTeleZone = m_hTeleZone.GetEntity();
                        CustomKeyvalues@ pTeleCustomKVs = pTeleZone.GetCustomKeyvalues();
                        CustomKeyvalue kvUnlocked(pTeleCustomKVs.GetKeyvalue("$i_unlocked"));
                        int unlocked = kvUnlocked.GetInteger();
                        pTeleCustomKVs.SetKeyvalue("$i_unlocked", unlocked + 1);

                        pCustomKVs.SetKeyvalue("$i_discovered", 1);
                        g_SoundSystem.EmitAmbientSound(pLung.edict(), pLung.pev.origin, "buttons/bell1.wav", 1.0, ATTN_NONE, 0, PITCH_NORM);
                        g_EntityFuncs.FireTargets(artifactTarget, null, null, USE_TOGGLE);
                        return;
                    }
                }
            }
        }

        void PlayFaunaNoise()
        {
            if (!m_hDepthIndicator.IsValid())
            {
                g_Game.AlertMessage(at_error, "Ragemap2024::Erty::MapPart->PlayFaunaNoise(): Depth indicator not found\n");
                return;
            }
            CBaseEntity@ pDepthIndicator = m_hDepthIndicator.GetEntity();

            int index = Math.RandomLong(0, peels.length() - 1);
            g_SoundSystem.EmitAmbientSound(pDepthIndicator.edict(), pDepthIndicator.pev.origin, peels[index], 1.0, ATTN_NONE, 0, PITCH_NORM);
            
            return;
        }

        void StopSinking()
        {
            m_fLungSinkSpeed = 0.0;
            if (!m_hLung.IsValid())
            {
                g_Game.AlertMessage(at_error, "Ragemap2024::Erty::MapPart->StopSinking(): Lung not found\n");
                return;
            }
            CBaseEntity@ pLung = m_hLung.GetEntity();
            pLung.pev.origin.z = m_fLungInitialDepth;
            return;
        }
    }

    MapPart@ g_pInstance;


    /*
     * -------------------------------------------------------------------------
     * Map hooks
     * -------------------------------------------------------------------------
     */
    
    /**
     * Map hook: Take Photo
     * @param  CBaseEntity@|null pActivator Activator entity
     * @param  CBaseEntity@|null pCaller    Caller entity
     * @param  USE_TYPE          useType    Use type, or unspecified to assume `USE_TOGGLE`
     * @param  float             flValue    Use value, or unspecified to assume `0.0f`
     * @return void
     */
    void TakePhoto(CBaseEntity@ pActivator, CBaseEntity@ pCaller, USE_TYPE useType, float flValue)
    {
        if (g_pInstance is null) {
            return;
        }

        g_pInstance.TakePhoto();
    }

    /**
     * Map hook: Update Lung Coordinates
     * @param  CBaseEntity@|null pActivator Activator entity
     * @return void
     */
    void UpdateLungCoords(CBaseEntity@ pTriggerScript)
    {
        if (g_pInstance is null) {
            return;
        }
        g_pInstance.UpdateLungCoords();
    }

    /**
     * Map hook: Play random fauna noises
     * @param  CBaseEntity@|null pActivator Activator entity
     * @param  CBaseEntity@|null pCaller    Caller entity
     * @param  USE_TYPE          useType    Use type, or unspecified to assume `USE_TOGGLE`
     * @param  float             flValue    Use value, or unspecified to assume `0.0f`
     * @return void
     */
    void PlayFaunaNoise(CBaseEntity@ pActivator, CBaseEntity@ pCaller, USE_TYPE useType, float flValue)
    {
        if (g_pInstance is null) {
            return;
        }

        g_pInstance.PlayFaunaNoise();
    }

    /**
     * Map hook: Sink or float the Lung
     * @param  CBaseEntity@|null pActivator Activator entity
     * @param  CBaseEntity@|null pCaller    Caller entity
     * @param  USE_TYPE          useType    Use type, or unspecified to assume `USE_TOGGLE`
     * @param  float             flValue    Use value, or unspecified to assume `0.0f`
     * @return void
     */
    void SinkOrFloat(CBaseEntity@ pActivator, CBaseEntity@ pCaller, USE_TYPE useType, float flValue)
    {
        if (g_pInstance is null) {
            return;
        }

        float amount = 4;
        if (pActivator.pev.dmg > 0)
        {
            amount = pActivator.pev.dmg + Math.RandomFloat(0, pActivator.pev.dmg / 2);
        }

        g_pInstance.SinkOrFloat(amount);
    }

    /**
     * Map hook: Stop sinking
     * @param  CBaseEntity@|null pActivator Activator entity
     * @param  CBaseEntity@|null pCaller    Caller entity
     * @param  USE_TYPE          useType    Use type, or unspecified to assume `USE_TOGGLE`
     * @param  float             flValue    Use value, or unspecified to assume `0.0f`
     * @return void
     */
    void StopSinking(CBaseEntity@ pActivator, CBaseEntity@ pCaller, USE_TYPE useType, float flValue)
    {
        if (g_pInstance is null) {
            return;
        }

        g_pInstance.StopSinking();
    }

}
