class func_erty_lung : ScriptBaseEntity
{
    bool m_isActive;
    float m_speed;
    float m_knockbackAmount;
    float m_shakeFrequency;
    float m_shakeDuration;
    float m_nextShake;
    array<string> m_damageSounds = {
        "debris/bustmetal1.wav",
        "debris/bustmetal2.wav",
        "debris/metal1.wav",
        "debris/metal2.wav",
        "debris/metal3.wav",
        "debris/metal4.wav",
        "debris/metal5.wav",
        "debris/metal6.wav",
        "debris/metal7.wav",
        "debris/impact_crete.wav"
    };

    void NextThink(float thinkTime, const bool alwaysThink)
	{
		if (alwaysThink) {
            self.pev.flags |= FL_ALWAYSTHINK;
        } else {
            self.pev.flags &= ~FL_ALWAYSTHINK;
        }
		self.pev.nextthink = thinkTime;
	}

	void Spawn()
	{
        m_isActive = false;
        m_knockbackAmount = 40.0;
        m_shakeFrequency = 42.0;
        m_shakeDuration = 0.6;
        m_nextShake = 0;

        Precache();

		if (self.pev.speed == 0) {
            m_speed = 32;
        } else {
            m_speed = self.pev.speed;
        }

        self.pev.max_health = 100.0;
        if (self.pev.health == 0.0)
        {
            self.pev.health = 100.0;
        }

		self.pev.speed = m_speed;
		self.pev.velocity = g_vecZero;
		self.pev.avelocity = g_vecZero;
		self.pev.impulse = int(m_speed);
		self.pev.solid = SOLID_TRIGGER; //SOLID_NOT;
		self.pev.movetype = MOVETYPE_FLY; //MOVETYPE_NOCLIP;
		g_EntityFuncs.SetSize(self.pev, Vector(-16, -16, -16), Vector(16, 16, 16));
        self.pev.takedamage = DAMAGE_YES;
        self.pev.renderamt = 0;
        self.pev.rendermode = kRenderTransTexture;
        self.pev.angles = g_vecZero;

        Math.MakeVectors(self.pev.angles);
        self.pev.movedir = g_Engine.v_forward;

		g_EntityFuncs.SetOrigin(self, self.pev.origin);
		self.pev.oldorigin = self.pev.origin;

        SetThink(ThinkFunction(this.LungThink));
        SetTouch(TouchFunction(this.LungTouch));
	}

    void Precache()
    {
        g_SoundSystem.PrecacheSound("debris/bustmetal1.wav");
        g_SoundSystem.PrecacheSound("debris/bustmetal2.wav");
        g_SoundSystem.PrecacheSound("debris/metal1.wav");
        g_SoundSystem.PrecacheSound("debris/metal2.wav");
        g_SoundSystem.PrecacheSound("debris/metal3.wav");
        g_SoundSystem.PrecacheSound("debris/metal4.wav");
        g_SoundSystem.PrecacheSound("debris/metal5.wav");
        g_SoundSystem.PrecacheSound("debris/metal6.wav");
        g_SoundSystem.PrecacheSound("debris/metal7.wav");
        g_SoundSystem.PrecacheSound("debris/impact_crete.wav");
    }

    void LungDeath()
    {
        g_EntityFuncs.FireTargets("erty_lung_dead_mm", null, null, USE_TOGGLE);
        self.pev.speed = 0.0;
        self.pev.velocity = g_vecZero;
        self.pev.flags &= ~FL_ALWAYSTHINK;
        SetThink(null);
    }

    void PlayDamageSoundAndShake(int amplitude)
    {
        int index = Math.RandomLong(0, m_damageSounds.length() - 1);
        int pitch = Math.RandomLong(52, 69);
        g_SoundSystem.EmitAmbientSound(self.edict(), self.pev.origin, m_damageSounds[index], 1.0, ATTN_NONE, 0, pitch);
        g_PlayerFuncs.ScreenShakeAll(self.pev.origin, amplitude + 5, m_shakeFrequency, m_shakeDuration);
    }

    void LungKnockback(CBaseEntity@ pInflictor)
    {
        int amplitude = Math.RandomLong(1, 10);

        if (self.pev.health - amplitude <= 0) {
            LungDeath();
        }
        else
        {
            self.TakeDamage(pInflictor.pev, pInflictor.pev, amplitude, DMG_CRUSH);
            PlayDamageSoundAndShake(amplitude);
            g_EntityFuncs.SetOrigin(self, self.pev.origin - (self.pev.movedir * m_knockbackAmount));
        }
    }

    void DepthCrush(entvars_t@ pInflictor)
    {
        if (g_Engine.time > m_nextShake)
        {
            int amplitude = Math.RandomLong(1, 4);
            if (self.pev.health - amplitude <= 0) {
                LungDeath();
            }
            else
            {
                BaseClass.TakeDamage(pInflictor, pInflictor, amplitude, DMG_CRUSH);
                PlayDamageSoundAndShake(Math.RandomLong(1, 3));
            }

            m_nextShake = g_Engine.time + Math.RandomFloat(0.3, 1.8);
        }
    }

    int TakeDamage(entvars_t@ pevInflictor, entvars_t@ pevAttacker, float flDamage, int bitsDamageType)
    {
        if (pevInflictor.targetname == "erty_noparking")
        {
            DepthCrush(pevInflictor);
            return 1;
        }
        return BaseClass.TakeDamage(pevInflictor, pevAttacker, flDamage, bitsDamageType);
    }

    void LungTouch(CBaseEntity@ pOther)
    {
        if (pOther.pev.classname == "worldspawn")
        {
            LungKnockback(pOther);
        }
    }

    void LungThink()
    {
        if (m_isActive)
        {
            Math.MakeVectors(self.pev.angles);
            self.pev.movedir = g_Engine.v_forward;
            self.pev.velocity = self.pev.speed * self.pev.movedir;

            NextThink(g_Engine.time + 0.1, true);
        }
    }

    void Use(CBaseEntity@ pActivator, CBaseEntity@ pCaller, USE_TYPE useType, float value)
    {
        if (useType == USE_TOGGLE) {
            m_isActive = !m_isActive;
        } else {
            m_isActive == bool(useType == USE_ON);
        }

        if (m_isActive)
        {
            NextThink(g_Engine.time + 0.1, true);
        }
        else
        {
            self.pev.velocity = g_vecZero;
            self.pev.flags &= ~FL_ALWAYSTHINK;
        }
    }

}
