// INS2 Flamethrower
#include "ins2/explo/weapon_ins2m2"

// Controllable Alien Slave
#include "controlnpc/cnpc"

namespace GIEGUE
{

EHandle g_hHurt;
EHandle g_hBoss;

int g_iPhase = 1;
bool g_bIntermission = false;

int g_iSpores = 0;

Vector g_vecFloor;
Vector g_vecVoid;

const int PHASE_HP_THRESHOLD = 100;
const int PHASE_HP_BASE = 375;

const int BOSS_CRUSH_SPEED = 640;

const int TICKETS_PER_PLAYER = 6; // starting tickets = [TICKETS_PER_PLAYER * num_of_players]
const int TICKETS_PER_PHASE = 3; // extra tickets given on checkpoints = [TICKETS_PER_PHASE * num_of_players]

void MapInit()
{
	// Register the flamethrower
	INS2_M2FLAMETHROWER::Register();
	
	// Controllable monster logic
	CNPC::Initialize();
	
	// Other custom entities
	g_CustomEntityFuncs.RegisterCustomEntity( "GIEGUE::CBounceDart", "bounce_dart" );
	
	// Sprite HUD for Tickets
	g_Game.PrecacheModel( "sprites/ragemap2023/tickets.spr" );
	
	// Boss attack sounds
	g_SoundSystem.PrecacheSound( "player/pl_long_jump.wav" );
	g_SoundSystem.PrecacheSound( "weapons/glauncher.wav" );
	g_SoundSystem.PrecacheSound( "weapons/glauncher2.wav" );
	g_SoundSystem.PrecacheSound( "ambience/rocket_steam1.wav" );
	g_SoundSystem.PrecacheSound( "misc/bbimpact.wav" );
	g_SoundSystem.PrecacheSound( "ragemap2020/ClockTick.wav" );
	
	// Attack sounds that need to be downloaded to players
	g_Game.PrecacheGeneric( "sound/ragemap2020/ClockTick.wav" );
	
	// Boss sprites
	g_Game.PrecacheModel( "sprites/stmbal1.spr" );
	g_Game.PrecacheModel( "sprites/laserbeam.spr" );
}

void PartStart( CBaseEntity@ pActivator, CBaseEntity@ pCaller, USE_TYPE useType, float flValue )
{
	// setup some map cfgs
	g_EngineFuncs.CVarSetFloat( "sv_maxspeed", 320 );
	g_EngineFuncs.CVarSetFloat( "sv_gravity", 720 );
	
	g_EngineFuncs.CVarSetFloat( "mp_suitpower", 0 );
	g_EngineFuncs.CVarSetFloat( "mp_observer_mode", 1 );
	g_EngineFuncs.CVarSetFloat( "mp_respawndelay", 2 );
	g_EngineFuncs.CVarSetFloat( "mp_dropweapons", 0 );
	g_EngineFuncs.CVarSetFloat( "mp_flashlight", 0 );
	
	for ( int i = 1; i <= g_Engine.maxClients; i++ )
	{
		CBasePlayer@ pPlayer = g_PlayerFuncs.FindPlayerByIndex( i );
		if ( pPlayer !is null && pPlayer.IsConnected() )
		{
			// little fix
			pPlayer.SetMaxSpeed( int( g_EngineFuncs.CVarGetFloat( "sv_maxspeed" ) ) );
			pPlayer.pev.targetname = "living_player";
		}
	}
	
	// setup the flamethrower
	INS2FLAMEPROJECTILE::ALLOW_EFFECTS = false;
	INS2_M2FLAMETHROWER::AMMO_INFINITE = true;
	
	// adjust some skill cvars
	g_EngineFuncs.CVarSetFloat( "sk_agrunt_health", 75 );
	g_EngineFuncs.CVarSetFloat( "sk_bigmomma_dmg_blast", 70 );
	g_EngineFuncs.CVarSetFloat( "sk_bigmomma_radius_blast", 140 );
	g_EngineFuncs.CVarSetFloat( "sk_agrunt_berserker_dmg_punch", 20 );
}

void PartEnd( CBaseEntity@ pActivator, CBaseEntity@ pCaller, USE_TYPE useType, float flValue )
{
	// turn off survival if active
	g_EngineFuncs.CVarSetFloat( "mp_observer_cyclic", 0 );
	SURVIVAL::HideTickets();
	
	// keep it off!
	g_EntityFuncs.FireTargets( "giegue_boss_spawn", null, null, USE_OFF, 0.0f, 0.0f );
	g_EntityFuncs.FireTargets( "giegue_boss_music", null, null, USE_OFF, 0.0f, 0.0f ); // thanks grunt
	
	// remove hooks
	g_Hooks.RemoveHook( Hooks::Player::ClientPutInServer );
	g_Hooks.RemoveHook( Hooks::Player::PlayerKilled );
	g_Hooks.RemoveHook( Hooks::Player::PlayerTakeDamage );
	g_Hooks.RemoveHook( Hooks::Player::PlayerPostThink );
	
	// restore flamethrower behaviour
	INS2FLAMEPROJECTILE::ALLOW_EFFECTS = true;
	INS2FLAMEPROJECTILE::SOLID_TYPE = SOLID_BBOX;
	INS2_M2FLAMETHROWER::AMMO_INFINITE = false;
	INS2_M2FLAMETHROWER::FIRST_DEPLOY = true;
	
	// unhide weapon HUDs
	for ( int i = 1; i <= g_Engine.maxClients; i++ )
	{
		CBasePlayer@ pPlayer = g_PlayerFuncs.FindPlayerByIndex( i );
		if ( pPlayer !is null && pPlayer.IsConnected() )
		{
			pPlayer.RemoveAllItems( false );
			pPlayer.m_iHideHUD = 0;
		}
	}
	
	// restore skill cvars
	g_EngineFuncs.CVarSetFloat( "sk_agrunt_health", 150 );
	g_EngineFuncs.CVarSetFloat( "sk_bigmomma_dmg_blast", 140 );
	g_EngineFuncs.CVarSetFloat( "sk_bigmomma_radius_blast", 260 );
	g_EngineFuncs.CVarSetFloat( "sk_agrunt_berserker_dmg_punch", 30 );
}

void BossInit( CBaseEntity@ pActivator, CBaseEntity@ pCaller, USE_TYPE useType, float flValue )
{
	g_hHurt = g_EntityFuncs.FindEntityByTargetname( null, "giegue_barrel_hurt" );
	g_hBoss = g_EntityFuncs.FindEntityByTargetname( null, "giegue_boss_barrel" );
	
	if ( g_hHurt.IsValid() && g_hBoss.IsValid() )
	{
		CBaseEntity@ pHurt = g_hHurt.GetEntity();
		CBaseEntity@ pBoss = g_hBoss.GetEntity();
		
		// setup barrel (func_breakable)
		pBoss.pev.nextthink = Math.FLOAT_MAX;
		
		// setup crusher (trigger_hurt)
		pHurt.pev.movetype = MOVETYPE_FOLLOW;
		@pHurt.pev.aiment = pBoss.edict();
		
		// setup boss health per phase
		pBoss.pev.health = ( PHASE_HP_BASE * g_PlayerFuncs.GetNumPlayers() ) + PHASE_HP_THRESHOLD;
		
		// locate floor Z origin
		CBaseEntity@ pLocator = g_EntityFuncs.FindEntityByTargetname( null, "giegue_floor_locator" );
		
		TraceResult trZ;
		g_Utility.TraceLine( pLocator.pev.origin, pLocator.pev.origin + Vector( 0, 0, -256 ), ignore_monsters, pLocator.edict(), trZ );
		
		g_vecFloor = trZ.vecEndPos + Vector( 0, 0, 1 );
		
		// locate void room origin
		@pLocator = g_EntityFuncs.FindEntityByTargetname( null, "giegue_void_locator" );
		g_vecVoid = pLocator.pev.origin;
		
		// activate the hooks
		g_Hooks.RegisterHook( Hooks::Player::ClientPutInServer, @ClientPutInServer );
		g_Hooks.RegisterHook( Hooks::Player::PlayerKilled, @PlayerKilled );
		g_Hooks.RegisterHook( Hooks::Player::PlayerTakeDamage, @PlayerTakeDamage );
		g_Hooks.RegisterHook( Hooks::Player::PlayerPostThink, @PlayerPostThink );
		g_Scheduler.SetTimeout( "HurtRemoveDelay", 0.01 );
		g_Scheduler.SetTimeout( "ScanBossHealth", 0.10 );
		g_Scheduler.SetTimeout( "PlayerHPRegen", 0.25 );
		SURVIVAL::Init();
		
		// give the weapon to the players
		for ( int i = 1; i <= g_Engine.maxClients; i++ )
		{
			CBasePlayer@ pPlayer = g_PlayerFuncs.FindPlayerByIndex( i );
			if ( pPlayer !is null && pPlayer.IsConnected() )
			{
				pPlayer.GiveNamedItem( "weapon_ins2m2", SF_NORESPAWN );
				pPlayer.m_iHideHUD = 1; // no constant for this? HIDEHUD_WEAPONS
			}
		}
		INS2_M2FLAMETHROWER::FIRST_DEPLOY = false;
		
		// very fast hint for those quick enough to read it
		g_PlayerFuncs.ClientPrintAll( HUD_PRINTCENTER, "HINT\nRed means danger! Get out of the way!\n" );
	}
}

void HurtRemoveDelay()
{
	// as long as the boss exists
	if ( g_hBoss.IsValid() && g_hHurt.IsValid() )
	{
		CBaseEntity@ pHurt = g_hHurt.GetEntity();
		pHurt.pev.dmgtime = g_Engine.time;
		
		g_Scheduler.SetTimeout( "HurtRemoveDelay", 0.01 );
	}
}

void ScanBossHealth()
{
	if ( g_hBoss.IsValid() )
	{
		CBaseEntity@ pBoss = g_hBoss.GetEntity();
		if ( !g_bIntermission && pBoss.pev.health <= PHASE_HP_THRESHOLD && g_iPhase < 7 )
		{
			g_bIntermission = true;
			g_iPhase++;
			
			// invulnerable during phase transition
			pBoss.pev.velocity = g_vecZero;
			pBoss.pev.takedamage = DAMAGE_NO;
			pBoss.pev.movetype = MOVETYPE_PUSH;
			
			// sound and sprite effects
			g_SoundSystem.EmitSoundDyn( pBoss.edict(), CHAN_BODY, "ambience/rocket_steam1.wav", VOL_NORM, ATTN_NORM, SND_FORCE_SINGLE, PITCH_NORM );
			
			for ( uint I = 1; I <= 4; I++ )
			{
				array< Vector > vecOrigin =
				{
					// hhhhhhhhhhh
					Vector( pBoss.pev.origin.x + ( pBoss.pev.size.x * ( -0.3 - ( float( I ) / 10.0 ) ) ), pBoss.pev.origin.y + ( pBoss.pev.size.y * ( 0.3 + ( float( I ) / 10.0 ) ) ), pBoss.pev.origin.z + ( pBoss.pev.size.z * 0.45 ) ),
					Vector( pBoss.pev.origin.x + ( pBoss.pev.size.x * ( 0.3 + ( float( I ) / 10.0 ) ) ), pBoss.pev.origin.y + ( pBoss.pev.size.y * ( -0.3 - ( float( I ) / 10.0 ) ) ), pBoss.pev.origin.z + ( pBoss.pev.size.z * 0.45 ) ),
					Vector( pBoss.pev.origin.x + ( pBoss.pev.size.x * ( -0.3 - ( float( I ) / 10.0 ) ) ), pBoss.pev.origin.y + ( pBoss.pev.size.y * ( -0.3 - ( float( I ) / 10.0 ) ) ), pBoss.pev.origin.z + ( pBoss.pev.size.z * 0.45 ) ),
					Vector( pBoss.pev.origin.x + ( pBoss.pev.size.x * ( 0.3 + ( float( I ) / 10.0 ) ) ), pBoss.pev.origin.y + ( pBoss.pev.size.y * ( 0.3 + ( float( I ) / 10.0 ) ) ), pBoss.pev.origin.z + ( pBoss.pev.size.z * 0.45 ) )
				};
				
				for ( uint J = 0; J < 4; J++ )
				{
					CSprite@ pSprite = g_EntityFuncs.CreateSprite( "sprites/stmbal1.spr", vecOrigin[ J ], false );
					pSprite.SetTransparency( kRenderTransAdd, 255, 255, 255, 250, kRenderFxNone );
					pSprite.pev.targetname = "giegue_temp_" + string( I ) + "_" + string( J );
					g_EntityFuncs.FireTargets( pSprite.pev.targetname, null, null, USE_OFF, 0.0f, 0.0f ); // could you please start OFF?
					g_EntityFuncs.FireTargets( pSprite.pev.targetname, null, null, USE_ON, 0.0f, 0.0f + ( float( I ) / 10.0 ) );
					g_EntityFuncs.FireTargets( pSprite.pev.targetname, null, null, USE_OFF, 0.0f, 1.0f + ( float( I ) / 10.0 ) );
				}
			}
			
			g_Scheduler.SetTimeout( "InitNextPhase", 1.3 );
		}
		else if ( !g_bIntermission && pBoss.pev.health <= PHASE_HP_THRESHOLD && g_iPhase >= 7 )
		{
			// paranoia calling again
			if ( g_iPhase == 7 )
			{
				// play sound and summon sprites but don't turn them off
				g_SoundSystem.EmitSoundDyn( pBoss.edict(), CHAN_BODY, "ambience/rocket_steam1.wav", VOL_NORM, ATTN_NORM, 0, PITCH_NORM );
				
				for ( uint I = 1; I <= 4; I++ )
				{
					array< Vector > vecOrigin =
					{
						// hhhhhhhhhhh
						Vector( pBoss.pev.origin.x + ( pBoss.pev.size.x * ( -0.3 - ( float( I ) / 10.0 ) ) ), pBoss.pev.origin.y + ( pBoss.pev.size.y * ( 0.3 + ( float( I ) / 10.0 ) ) ), pBoss.pev.origin.z + ( pBoss.pev.size.z * 0.45 ) ),
						Vector( pBoss.pev.origin.x + ( pBoss.pev.size.x * ( 0.3 + ( float( I ) / 10.0 ) ) ), pBoss.pev.origin.y + ( pBoss.pev.size.y * ( -0.3 - ( float( I ) / 10.0 ) ) ), pBoss.pev.origin.z + ( pBoss.pev.size.z * 0.45 ) ),
						Vector( pBoss.pev.origin.x + ( pBoss.pev.size.x * ( -0.3 - ( float( I ) / 10.0 ) ) ), pBoss.pev.origin.y + ( pBoss.pev.size.y * ( -0.3 - ( float( I ) / 10.0 ) ) ), pBoss.pev.origin.z + ( pBoss.pev.size.z * 0.45 ) ),
						Vector( pBoss.pev.origin.x + ( pBoss.pev.size.x * ( 0.3 + ( float( I ) / 10.0 ) ) ), pBoss.pev.origin.y + ( pBoss.pev.size.y * ( 0.3 + ( float( I ) / 10.0 ) ) ), pBoss.pev.origin.z + ( pBoss.pev.size.z * 0.45 ) )
					};
					
					for ( uint J = 0; J < 4; J++ )
					{
						CSprite@ pSprite = g_EntityFuncs.CreateSprite( "sprites/stmbal1.spr", vecOrigin[ J ], false );
						pSprite.SetTransparency( kRenderTransAdd, 255, 255, 255, 250, kRenderFxNone );
						pSprite.pev.targetname = "giegue_temp_" + string( I ) + "_" + string( J );
						g_EntityFuncs.FireTargets( pSprite.pev.targetname, null, null, USE_OFF, 0.0f, 0.0f ); // could you please start OFF?
						g_EntityFuncs.FireTargets( pSprite.pev.targetname, null, null, USE_ON, 0.0f, 0.0f + ( float( I ) / 10.0 ) );
					}
				}
				
				g_iPhase++;
				pBoss.pev.health = PHASE_HP_THRESHOLD * ( g_PlayerFuncs.GetNumPlayers() + 1 );
			}
			else
			{
				g_bIntermission = true;
				
				// ded
				pBoss.pev.takedamage = DAMAGE_NO;
				pBoss.pev.health = 1;
				
				// stop sound
				g_SoundSystem.StopSound( pBoss.edict(), CHAN_BODY, "ambience/rocket_steam1.wav" );
				
				// now do turn off the sprites
				for ( uint I = 1; I <= 4; I++ )
				{
					for ( uint J = 0; J < 4; J++ )
					{
						string szSpriteName = "giegue_temp_" + string( I ) + "_" + string( J );
						g_EntityFuncs.FireTargets( szSpriteName, null, null, USE_OFF, 0.0f, 0.0f + ( float( I ) / 10.0 ) );
					}
				}
				
				// EXPLOSION! (with all caps)
				BarrelDeath();
			}
		}
		
		g_Scheduler.SetTimeout( "ScanBossHealth", 0.10 );
	}
}

void InitNextPhase()
{
	if ( g_hBoss.IsValid() )
	{
		CBaseEntity@ pBoss = g_hBoss.GetEntity();
		
		// stop previous sound
		g_SoundSystem.StopSound( pBoss.edict(), CHAN_BODY, "ambience/rocket_steam1.wav" );
		
		pBoss.pev.health = PHASE_HP_THRESHOLD + 1;
		g_bIntermission = false;
		
		Vector vecDest = Vector( pBoss.pev.origin.x, pBoss.pev.origin.y, 450 );
		LinearMove( pBoss, vecDest, BOSS_CRUSH_SPEED, "NextPhase", 0.5 );
	}
}

void NextPhase()
{
	if ( g_hBoss.IsValid() )
	{
		CBaseEntity@ pBoss = g_hBoss.GetEntity();
		
		// reset boss health
		pBoss.pev.health = ( PHASE_HP_BASE * g_PlayerFuncs.GetNumPlayers() ) + PHASE_HP_THRESHOLD;
		
		// remove sprites
		CBaseEntity@ pSprite = null;
		while ( ( @pSprite = g_EntityFuncs.FindEntityByClassname( pSprite, "env_sprite" ) ) !is null )
		{
			if ( string( pSprite.pev.targetname ).StartsWith( "giegue_temp_" ) )
				g_EntityFuncs.Remove( pSprite );
		}
		
		// get rid of monster add-ons
		CBaseEntity@ pMonster = null;
		while ( ( @pMonster = g_EntityFuncs.FindEntityByTargetname( pMonster, "giegue_minion_npc" ) ) !is null )
		{
			if ( pMonster.IsAlive() )
				pMonster.Killed( pBoss.pev, GIB_ALWAYS );
		}
		
		switch ( g_iPhase )
		{
			case 2: g_EntityFuncs.FireTargets( "giegue_boss_phase2", null, null, USE_TOGGLE, 0.0f, 0.0f ); break;
			case 3: PreludePhase3(); break;
			//case 4: g_EntityFuncs.FireTargets( "giegue_boss_phase4", null, null, USE_TOGGLE, 0.0f, 0.0f ); break;
			case 5: g_EntityFuncs.FireTargets( "giegue_boss_phase5", null, null, USE_TOGGLE, 0.0f, 0.0f ); break;
			//case 6: g_EntityFuncs.FireTargets( "giegue_boss_phase6", null, null, USE_TOGGLE, 0.0f, 0.0f ); break;
			case 7: BarrelFinish(); break;
		}
	}
}

void PreludePhase2( CBaseEntity@ pActivator, CBaseEntity@ pCaller, USE_TYPE useType, float flValue )
{
	if ( g_hBoss.IsValid() )
	{
		CBaseEntity@ pBoss = g_hBoss.GetEntity();
		
		// HP is too high for this phase, tone it down a bit
		pBoss.pev.health -= PHASE_HP_THRESHOLD * g_PlayerFuncs.GetNumPlayers();
		
		GetEnemy();
		g_Scheduler.SetTimeout( "BounceAttack", 0.9 );
	}
}

void PreludePhase3()
{
	// this whole thing could have been coded better X_X
	CBaseEntity@ pCountTrigger;
	
	// choose a random trigger to be the correct answer
	switch ( Math.RandomLong( 1, 3 ) )
	{
		case 1: // LEFT
		{
			@pCountTrigger = g_EntityFuncs.FindEntityByTargetname( null, "giegue_math_left" );
			pCountTrigger.KeyValue( "intarget", "dummy" );
			pCountTrigger.KeyValue( "incount", "dummy" );
			
			pCountTrigger.pev.iuser4 = g_iSpores;
			g_EntityFuncs.FindEntityByTargetname( null, "giegue_wcount_color1" ).pev.rendercolor = Vector( 0, 255, 0 );
			
			@pCountTrigger = g_EntityFuncs.FindEntityByTargetname( null, "giegue_math_middle" );
			pCountTrigger.pev.iuser4 = g_iSpores + 1;
			
			@pCountTrigger = g_EntityFuncs.FindEntityByTargetname( null, "giegue_math_right" );
			pCountTrigger.pev.iuser4 = g_iSpores + 2;
			
			break;
		}
		case 2: // MIDDLE
		{
			@pCountTrigger = g_EntityFuncs.FindEntityByTargetname( null, "giegue_math_middle" );
			pCountTrigger.KeyValue( "intarget", "dummy" );
			pCountTrigger.KeyValue( "incount", "dummy" );
			
			pCountTrigger.pev.iuser4 = g_iSpores;
			g_EntityFuncs.FindEntityByTargetname( null, "giegue_wcount_color2" ).pev.rendercolor = Vector( 0, 255, 0 );
			
			@pCountTrigger = g_EntityFuncs.FindEntityByTargetname( null, "giegue_math_left" );
			pCountTrigger.pev.iuser4 = g_iSpores - 1;
			
			@pCountTrigger = g_EntityFuncs.FindEntityByTargetname( null, "giegue_math_right" );
			pCountTrigger.pev.iuser4 = g_iSpores + 1;
			
			break;
		}
		case 3: // RIGHT
		{
			@pCountTrigger = g_EntityFuncs.FindEntityByTargetname( null, "giegue_math_right" );
			pCountTrigger.KeyValue( "intarget", "dummy" );
			pCountTrigger.KeyValue( "incount", "dummy" );
			
			pCountTrigger.pev.iuser4 = g_iSpores;
			g_EntityFuncs.FindEntityByTargetname( null, "giegue_wcount_color3" ).pev.rendercolor = Vector( 0, 255, 0 );
			
			@pCountTrigger = g_EntityFuncs.FindEntityByTargetname( null, "giegue_math_middle" );
			pCountTrigger.pev.iuser4 = g_iSpores - 1;
			
			@pCountTrigger = g_EntityFuncs.FindEntityByTargetname( null, "giegue_math_left" );
			pCountTrigger.pev.iuser4 = g_iSpores - 2;
			
			break;
		}
	}
	
	// setup the sprite numbers
	for ( uint I = 1; I <= 3; I++ )
	{
		switch ( I )
		{
			case 1: @pCountTrigger = g_EntityFuncs.FindEntityByTargetname( null, "giegue_math_left" ); break;
			case 2: @pCountTrigger = g_EntityFuncs.FindEntityByTargetname( null, "giegue_math_middle" ); break;
			case 3: @pCountTrigger = g_EntityFuncs.FindEntityByTargetname( null, "giegue_math_right" ); break;
		}
		CSprite@ pSprite = cast< CSprite@ >( g_EntityFuncs.FindEntityByTargetname( null, "giegue_wcount_spr" + string( I ) ) );
		pSprite.pev.framerate = 0;
		pSprite.pev.frame = 90 - pCountTrigger.pev.iuser4;
	}
	
	// gradually spawn the lasers
	CountSpawnLaser( 0 );
	
	// restore flame projectile before things get ugly
	INS2FLAMEPROJECTILE::SOLID_TYPE = SOLID_BBOX;
}
void CountSpawnLaser( int iStripe )
{
	Vector vecOrigin = g_vecFloor;
	
	vecOrigin.y -= -416 + ( 48 * iStripe );
	
	CBeam@ pBeam = null;
	for ( uint I = 0; I < 2; I++ )
	{
		@pBeam = g_EntityFuncs.CreateBeam( "sprites/laserbeam.spr", 4 );
		pBeam.PointsInit( vecOrigin + Vector( I == 0 ? -152 : 152, 0, 0 ), vecOrigin + Vector( I == 0 ? -152 : 152, 32, 0 ) );
		pBeam.SetColor( 200, 200, 200 );
		pBeam.SetBrightness( 225 );
		pBeam.RelinkBeam();
		
		@pBeam = g_EntityFuncs.CreateBeam( "sprites/laserbeam.spr", 4 );
		pBeam.PointsInit( vecOrigin + Vector( I == 0 ? -152 : 152, 0, 90 ), vecOrigin + Vector( I == 0 ? -152 : 152, 32, 90 ) );
		pBeam.SetColor( 200, 200, 200 );
		pBeam.SetBrightness( 225 );
		pBeam.RelinkBeam();
	}
	
	if ( iStripe < 18 )
	{
		iStripe++;
		g_Scheduler.SetTimeout( "CountSpawnLaser", 0.1, iStripe );
	}
	else
	{
		g_EntityFuncs.FireTargets( "giegue_wcount_display", null, null, USE_ON, 0.0f, 0.1f );
		g_Scheduler.SetTimeout( "CountAttack", 0.2 );
	}
}

void PreludePhase4( CBaseEntity@ pActivator, CBaseEntity@ pCaller, USE_TYPE useType, float flValue )
{
	if ( g_hBoss.IsValid() )
	{
		CBaseEntity@ pBoss = g_hBoss.GetEntity();
		
		// get rid of the lasers
		CBaseEntity@ pBeam = null;
		while ( ( @pBeam = g_EntityFuncs.FindEntityInSphere( pBeam, g_vecFloor, 1024.0, "beam", "classname" ) ) !is null )
		{
			g_EntityFuncs.Remove( pBeam );
		}
		
		g_iPhase = 4;
		
		// the barrel is still moving, keep HP a bit lower
		pBoss.pev.health -= PHASE_HP_THRESHOLD * g_PlayerFuncs.GetNumPlayers();
		
		// quickie hintie
		g_PlayerFuncs.ClientPrintAll( HUD_PRINTCENTER, "HINT\nYou can high jump while green!\n" );
	}
}
void LightJumpRotate( CBaseEntity@ pActivator, CBaseEntity@ pCaller, USE_TYPE useType, float flValue )
{
	if ( g_hBoss.IsValid() )
	{
		CBaseEntity@ pBoss = g_hBoss.GetEntity();
		
		CBaseEntity@ pLight = null;
		
		// setup solid/movetype to allow rotation
		@pLight = g_EntityFuncs.FindEntityByTargetname( null, "giegue_disco_pole" );
		pLight.pev.solid = SOLID_NOT;
		pLight.pev.movetype = MOVETYPE_NOCLIP;
		pLight.pev.avelocity = Vector( 0, 65, 0 );
		
		@pLight = g_EntityFuncs.FindEntityByTargetname( null, "giegue_disco_wlight" );
		pLight.pev.solid = SOLID_NOT;
		pLight.pev.movetype = MOVETYPE_NOCLIP;
		pLight.pev.avelocity = Vector( 0, 65, 0 );
		
		g_Scheduler.SetTimeout( "LightJumpScan", 0.1, EHandle( pLight ) );
	}
}
void LightJumpScan( EHandle hLight )
{
	// keep track of players
	for ( int iPlayerIndex = 1; iPlayerIndex <= g_Engine.maxClients; iPlayerIndex++ )
	{
		CBasePlayer@ pPlayer = g_PlayerFuncs.FindPlayerByIndex( iPlayerIndex );
		if ( pPlayer !is null && pPlayer.IsConnected() )
		{
			//if ( g_Utility.IsPlayerInVolume( pPlayer, g_EntityFuncs.FindEntityByTargetname( null, "giegue_disco_wlight" ) ) )
			//	g_PlayerFuncs.ClientPrint( pPlayer, HUD_PRINTTALK, "* DEBUG: Touch!\n" );
			
			if ( !hLight.IsValid() || g_Engine.time > float( pPlayer.GetUserData( "flJumpTime" ) ) )
			{
				// LightJumpOFF()
				pPlayer.pev.rendercolor.y = 50; // G
				if ( pPlayer.pev.rendercolor.x == 50 ) // R
					pPlayer.pev.renderfx = kRenderFxNone;
			}
		}
	}
	
	if ( hLight.IsValid() )
	{
		CBaseEntity@ pLight = hLight.GetEntity();
		
		float flAngle = pLight.pev.angles.y % 360;
		
		// keep track of entities
		CBaseEntity@ pTrigger = null;
		for ( int I = 1; I <= 12; I++ )
		{
			@pTrigger = g_EntityFuncs.FindEntityByTargetname( null, "giegue_disco_wtrigger" + string( I ) );
			pTrigger.pev.target = "dummy";
			
			if ( flAngle >= ( 15 * ( I - 1 ) ) && flAngle < ( 15 * I ) || flAngle >= 180 + ( 15 * ( I - 1 ) ) && flAngle < 180 + ( 15 * I ) )
				pTrigger.pev.target = "giegue_light_jump_on";
		}
		
		g_Scheduler.SetTimeout( "LightJumpScan", 0.1, EHandle( pLight ) );
	}
}
void LightJumpON( CBaseEntity@ pActivator, CBaseEntity@ pCaller, USE_TYPE useType, float flValue )
{
	if ( pActivator.IsPlayer() && pActivator.IsAlive() )
	{
		pActivator.pev.rendercolor.y = 200; // G
		pActivator.GetUserData( "flJumpTime" ) = g_Engine.time + 0.2;
		
		if ( pActivator.pev.renderfx == kRenderFxNone )
		{
			pActivator.pev.renderfx = kRenderFxGlowShell;
			pActivator.pev.rendercolor.x = 50; // R
			pActivator.pev.rendercolor.z = 50; // B
			pActivator.pev.renderamt = 1;
		}
	}
}

void InitFloorAttack( CBaseEntity@ pActivator, CBaseEntity@ pCaller, USE_TYPE useType, float flValue )
{
	if ( g_hBoss.IsValid() )
	{
		CBaseEntity@ pBoss = g_hBoss.GetEntity();
		
		GetEnemy();
		g_Scheduler.SetTimeout( "DoFloorAttack", 0.8 );
	}
}

void PreludePhase5( CBaseEntity@ pActivator, CBaseEntity@ pCaller, USE_TYPE useType, float flValue )
{
	if ( g_hBoss.IsValid() )
	{
		CBaseEntity@ pBoss = g_hBoss.GetEntity();
		
		g_EntityFuncs.SetOrigin( pBoss, Vector( g_vecFloor.x, g_vecFloor.y, 450 ) );
		
		Vector vecDest = Vector( pBoss.pev.origin.x, pBoss.pev.origin.y, g_vecFloor.z + ( pBoss.pev.size.z * 2 ) ); // mid-air
		LinearMove( pBoss, vecDest, BOSS_CRUSH_SPEED / 2, "TransformPlayerEffect", 1.0 );
	}
}
void TransformPlayerEffect()
{
	if ( g_hBoss.IsValid() )
	{
		CBaseEntity@ pBoss = g_hBoss.GetEntity();
		
		Vector vecOrigin = pBoss.pev.origin;
		vecOrigin.z -= ( pBoss.pev.size.z / 2 );
		
		Vector vecAngles = Vector( 0, Math.RandomLong( -180, 180 ), 0 );
		
		CBaseMonster@ pMonster = cast< CBaseMonster@ >( g_EntityFuncs.Create( "monster_generic", g_vecVoid, vecAngles, true ) );
		g_EntityFuncs.SetModel( pMonster, "models/Tor.mdl" );
		//pMonster.KeyValue( "disableai", "1" );
		pMonster.pev.spawnflags = 4; // Not solid
		g_EntityFuncs.DispatchSpawn( pMonster.edict() );
		
		pMonster.pev.targetname = "giegue_tor_npc";
		pMonster.StartMonster();
		pMonster.pev.nextthink = g_Engine.time;
		@pMonster.pev.owner = pBoss.edict();
		
		g_EntityFuncs.SetOrigin( pMonster, vecOrigin );
		
		g_EntityFuncs.FireTargets( "giegue_tor_animation", null, null, USE_TOGGLE, 0.0f, 1.0f );
		g_EntityFuncs.FireTargets( "giegue_tor_sound", null, null, USE_TOGGLE, 0.0f, 1.0f );
		g_Scheduler.SetTimeout( "TransformPlayerBeam", 1.0, EHandle( pMonster ) );
	}
}
void TransformPlayerBeam( EHandle hMonster )
{
	if ( hMonster.IsValid() )
	{
		CBaseEntity@ pMonster = hMonster.GetEntity();
		
		CBeam@ pBeam = null;
		for ( int iPlayerIndex = 1; iPlayerIndex <= g_Engine.maxClients; iPlayerIndex++ )
		{
			CBasePlayer@ pPlayer = g_PlayerFuncs.FindPlayerByIndex( iPlayerIndex );
			if ( pPlayer !is null && pPlayer.IsConnected() )
			{
				if ( pPlayer.IsAlive() )
				{
					@pBeam = g_EntityFuncs.CreateBeam( "sprites/lgtning.spr", 50 );
					pBeam.PointEntInit( pMonster.pev.origin + Vector( 0, 0, pMonster.pev.size.z / 2 ), pPlayer );
					pBeam.SetColor( 96, 128, 16 );
					pBeam.SetBrightness( 150 );
					pBeam.SetNoise( 80 );
					pBeam.SetScrollRate( 50 );
					pBeam.RelinkBeam();
				}
			}
		}
		
		g_Scheduler.SetTimeout( "TransformPlayerComplete", 2.0, EHandle( pMonster ) );
	}
}
void TransformPlayerComplete( EHandle hMonster )
{
	if ( hMonster.IsValid() )
	{
		CBaseEntity@ pMonster = hMonster.GetEntity();
		g_EntityFuncs.Remove( pMonster );
		
		g_EntityFuncs.FireTargets( "giegue_transform_fade", null, null, USE_TOGGLE, 0.0f, 0.0f );
		
		// get rid of the lasers
		CBaseEntity@ pBeam = null;
		while ( ( @pBeam = g_EntityFuncs.FindEntityInSphere( pBeam, g_vecFloor, 1024.0, "beam", "classname" ) ) !is null )
		{
			g_EntityFuncs.Remove( pBeam );
		}
		
		for ( int iPlayerIndex = 1; iPlayerIndex <= g_Engine.maxClients; iPlayerIndex++ )
		{
			CBasePlayer@ pPlayer = g_PlayerFuncs.FindPlayerByIndex( iPlayerIndex );
			if ( pPlayer !is null && pPlayer.IsConnected() )
			{
				if ( pPlayer.IsAlive() )
				{
					pPlayer.GiveNamedItem( "weapon_islave", SF_NORESPAWN );
				}
			}
		}
		
		g_Scheduler.SetTimeout( "DoTransformAttack", 1.0 );
	}
}

void PreludePhase6()
{
	// i just needed a delay.
	g_iPhase++;
	g_EntityFuncs.FireTargets( "giegue_boss_phase6", null, null, USE_TOGGLE, 0.0f, 0.0f );
}

void LaserPrepare( CBaseEntity@ pActivator, CBaseEntity@ pCaller, USE_TYPE useType, float flValue )
{
	CBaseEntity@ pTrigger = g_EntityFuncs.FindEntityByTargetname( null, "giegue_laser_transform" );
	if ( pTrigger !is null )
	{
		//pTrigger.pev.target = "giegue_laser_in";
		
		// this disables touch, manual scan, AGAIN.
		pTrigger.pev.nextthink = Math.FLOAT_MAX;
		pTrigger.pev.movetype = MOVETYPE_PUSH;
		pTrigger.pev.solid = SOLID_NOT; // for some bizzare reason it is pushing some players, fully disable it.
		g_Scheduler.SetTimeout( "ScanLaserTouch", 0.02, EHandle( pTrigger ) );
		
		pTrigger.pev.effects &= ~EF_NODRAW;
		pTrigger.pev.rendermode = kRenderTransColor;
		pTrigger.pev.rendercolor = Vector( 50, 200, 50 );
		pTrigger.pev.renderamt = 200;
		
		LinearMove( pTrigger, Vector( g_vecFloor.x + 528, pTrigger.pev.origin.y, pTrigger.pev.origin.z ), BOSS_CRUSH_SPEED * 2, "LaserReturn", 1.0 );
	}
}
void LaserReturn()
{
	CBaseEntity@ pTrigger = g_EntityFuncs.FindEntityByTargetname( null, "giegue_laser_transform" );
	if ( pTrigger !is null )
	{
		//pTrigger.pev.target = "giegue_laser_out";
		pTrigger.pev.rendercolor = Vector( 50, 200, 250 );
		
		LinearMove( pTrigger, Vector( g_vecFloor.x - 528, pTrigger.pev.origin.y, pTrigger.pev.origin.z ), BOSS_CRUSH_SPEED * 2, "LaserDestroy" );
	}
}
void LaserDestroy()
{
	CBaseEntity@ pTrigger = g_EntityFuncs.FindEntityByTargetname( null, "giegue_laser_transform" );
	if ( pTrigger !is null )
		g_EntityFuncs.Remove( pTrigger );
	
	g_Scheduler.SetTimeout( "NextAttack", 1.0 );
}
void ScanLaserTouch( EHandle hTrigger )
{
	if ( hTrigger.IsValid() )
	{
		CBaseEntity@ pTrigger = hTrigger.GetEntity();
		
		CBaseEntity@ pOther;
		@pOther = null; while ( ( @pOther = g_EntityFuncs.FindEntityByClassname( pOther, "player" ) ) !is null )
		{
			if ( pOther.IsAlive() && pTrigger.Intersects( pOther ) )
			{
				g_EntityFuncs.FireTargets( pTrigger.pev.rendercolor.z == 50 ? "giegue_laser_in" : "giegue_laser_out", pOther, null, USE_TOGGLE, 0.0f, 0.0f );
			}
		}
		@pOther = null; while ( ( @pOther = g_EntityFuncs.FindEntityByClassname( pOther, "monster_alien_slave" ) ) !is null )
		{
			if ( pOther.IsAlive() && pTrigger.Intersects( pOther ) && pTrigger.pev.rendercolor.z == 250 )
			{
				g_EntityFuncs.FireTargets( "giegue_laser_out", pOther, null, USE_TOGGLE, 0.0f, 0.0f );
			}
		}
		
		g_Scheduler.SetTimeout( "ScanLaserTouch", 0.02, EHandle( pTrigger ) );
	}
}
void LaserTransformA( CBaseEntity@ pActivator, CBaseEntity@ pCaller, USE_TYPE useType, float flValue )
{
	if ( pActivator.IsPlayer() )
	{
		CBasePlayer@ pPlayer = cast< CBasePlayer@ >( pActivator );
		if ( pPlayer.pev.effects & EF_NODRAW != 0 )
		{
			// revert transformation, making a copy alien slave
			CBaseEntity@ pWeapon = pPlayer.HasNamedPlayerItem( "weapon_islave" );
			CBaseEntity@ pMonster = FindEntityByOwner( null, "cnpc_islave", pPlayer.edict() );
			
			Vector vecOrigin = pMonster.pev.origin;
			Vector vecAngles = pMonster.pev.angles;
			
			g_EntityFuncs.Remove( pWeapon );
			g_EntityFuncs.Remove( pMonster );
			
			pPlayer.pev.flags |= FL_NOTARGET; // don't attack the player as soon as you spawn
			CBaseEntity@ pCopy = g_EntityFuncs.Create( "monster_alien_slave", vecOrigin, vecAngles, true );
			pCopy.KeyValue( "classify", "-1" );
			pCopy.KeyValue( "displayname", string( pPlayer.pev.netname ) );
			g_EntityFuncs.DispatchSpawn( pCopy.edict() );
			@pCopy.pev.owner = pPlayer.edict();
			pCopy.pev.nextthink = -1; // wait
			
			// don't let the player attack just yet
			pPlayer.m_iEffectBlockWeapons = 1;
			pPlayer.ApplyEffects();
			pPlayer.pev.effects &= ~EF_NODRAW;
		}
	}
}
void LaserTransformB( CBaseEntity@ pActivator, CBaseEntity@ pCaller, USE_TYPE useType, float flValue )
{
	// attack
	if ( pActivator.IsMonster() && !pActivator.IsPlayer() ) // players are monsters, kinda
	{
		pActivator.KeyValue( "classify", "7" );
		@pActivator.pev.owner = null;
		pActivator.pev.nextthink = g_Engine.time;
	}
	else if ( pActivator.IsPlayer() )
	{
		CBasePlayer@ pPlayer = cast< CBasePlayer@ >( pActivator );
		
		pPlayer.m_iEffectBlockWeapons = 0;
		pPlayer.pev.flags &= ~FL_NOTARGET;
		pPlayer.ApplyEffects();
	}
}

void BarrelFinish()
{
	if ( g_hBoss.IsValid() )
	{
		CBaseEntity@ pBoss = g_hBoss.GetEntity();
		g_EntityFuncs.SetOrigin( pBoss, Vector( g_vecFloor.x, g_vecFloor.y, 450 ) );
		
		g_bIntermission = false;
		
		// slow peek
		Vector vecDest = Vector( g_vecFloor.x, g_vecFloor.y, 315 );
		LinearMove( pBoss, vecDest, BOSS_CRUSH_SPEED / 4, "BarrelShake", 1.0 );
	}
}
void BarrelShake()
{
	if ( g_hBoss.IsValid() )
	{
		CBaseEntity@ pBoss = g_hBoss.GetEntity();
		
		int iShake = int( pBoss.GetUserData( "shake" ) );
		
		iShake++;
		pBoss.GetUserData( "shake" ) = iShake;
		
		if ( iShake < 40 )
		{
			Vector vecDest = Vector( g_vecFloor.x + Math.RandomFloat( -8, 8 ), g_vecFloor.y + Math.RandomFloat( -8, 8 ), 315 );
			LinearMove( pBoss, vecDest, BOSS_CRUSH_SPEED * 2, "BarrelShake", 0.0 );
		}
		else
		{
			Vector vecDest = Vector( g_vecFloor.x, g_vecFloor.y, 315 );
			LinearMove( pBoss, vecDest, BOSS_CRUSH_SPEED, "BarrelDefeat", 0.0 );
		}
	}
}
void BarrelDefeat()
{
	if ( g_hBoss.IsValid() )
	{
		CBaseEntity@ pBoss = g_hBoss.GetEntity();
		
		pBoss.pev.takedamage = DAMAGE_YES;
		pBoss.pev.movetype = MOVETYPE_PUSHSTEP;
		INS2FLAMEPROJECTILE::SOLID_TYPE = SOLID_NOT; // you know the drill...
		ScanFlameHit();
		
		// get it a bit lower
		pBoss.pev.health -= PHASE_HP_THRESHOLD * g_PlayerFuncs.GetNumPlayers();
		
		// ambient_music lacks the fadeout that ambient_generic has. simulate one
		CBaseEntity@ pMusic = g_EntityFuncs.FindEntityByTargetname( null, "giegue_boss_music" );
		g_Scheduler.SetTimeout( "MusicFadeout", 0.1, EHandle( pMusic ), 1.0f );
	}
}

void MusicFadeout( EHandle hMusic, float flVolume )
{
	if ( hMusic.IsValid() )
	{
		CBaseEntity@ pMusic = hMusic.GetEntity();
		
		flVolume -= 0.025;
		if ( flVolume < 0.025 )
			g_EntityFuncs.Remove( pMusic ); // don't need it anymore
		else
		{
			g_SoundSystem.EmitSoundDyn( pMusic.edict(), CHAN_MUSIC, "ragemap2024/giegue/bgm_boss.ogg", flVolume, ATTN_NONE, SND_CHANGE_VOL, PITCH_NORM );
			g_Scheduler.SetTimeout( "MusicFadeout", 0.1, EHandle( pMusic ), flVolume );
		}
	}
}

void BarrelDeath()
{
	if ( g_hBoss.IsValid() )
	{
		CBaseEntity@ pBoss = g_hBoss.GetEntity();
		
		int iTimes = int( pBoss.GetUserData( "death_counter" ) );
		
		iTimes++;
		pBoss.GetUserData( "death_counter" ) = iTimes;
		
		Vector vecOrigin = pBoss.pev.origin;
		vecOrigin.z += pBoss.pev.size.z * 0.45;
		
		vecOrigin.x += Math.RandomFloat( -( pBoss.pev.size.x * 0.35 ), ( pBoss.pev.size.x * 0.35 ) );
		vecOrigin.y += Math.RandomFloat( -( pBoss.pev.size.y * 0.35 ), ( pBoss.pev.size.y * 0.35 ) );
		
		g_EntityFuncs.CreateExplosion( vecOrigin, Vector( -90, 0, 0 ), pBoss.edict(), Math.RandomLong( 80, 120 ), false );
		
		if ( iTimes >= 15 )
			g_EntityFuncs.FireTargets( "giegue_boss_shake", null, null, USE_TOGGLE, 0.0f, 0.0f );
		
		if ( iTimes < 25 )
			g_Scheduler.SetTimeout( "BarrelDeath", 0.14 );
		else
		{
			g_EntityFuncs.FireTargets( "giegue_bosswin_fade", null, null, USE_TOGGLE, 0.0f, 0.0f );
			
			pBoss.pev.takedamage = DAMAGE_YES;
			g_EntityFuncs.FireTargets( "giegue_boss_barrel", null, null, USE_TOGGLE, 0.0f, 0.25f );
		}
	}
}

void PlayerHPRegen()
{
	if ( g_hBoss.IsValid() )
	{
		for ( int iPlayerIndex = 1; iPlayerIndex <= g_Engine.maxClients; iPlayerIndex++ )
		{
			CBasePlayer@ pPlayer = g_PlayerFuncs.FindPlayerByIndex( iPlayerIndex );
			if ( pPlayer !is null && pPlayer.IsConnected() )
			{
				if ( pPlayer.IsAlive() && pPlayer.pev.health < pPlayer.pev.max_health )
				{
					if ( g_Engine.time > float( pPlayer.GetUserData( "flNextRegenTime" ) ) )
					{
						pPlayer.pev.health = pPlayer.pev.health + 1.0; // paranoia
						pPlayer.GetUserData( "flNextRegenTime" ) = g_Engine.time + 0.5;
					}
				}
			}
		}
		g_Scheduler.SetTimeout( "PlayerHPRegen", 0.25 );
	}
}

HookReturnCode ClientPutInServer( CBasePlayer@ pPlayer )
{
	// respawn after 10 extra seconds to avoid ticket wasting during player loading
	g_Scheduler.SetTimeout( "ClientPutInServerPOST", 10.0, EHandle( pPlayer ) );
	return HOOK_CONTINUE;
}
void ClientPutInServerPOST( EHandle hPlayer )
{
	if ( hPlayer.IsValid() )
	{
		CBasePlayer@ pPlayer = cast< CBasePlayer@ >( hPlayer.GetEntity() );
		
		if ( SURVIVAL::g_bActive )
			SURVIVAL::ShowTickets( pPlayer );
		
		// if survival is already active, avoid respawning a joining player
		if ( !SURVIVAL::IsEnabled() )
			PlayerRespawn( EHandle( pPlayer ) );
	}
}

HookReturnCode PlayerTakeDamage( DamageInfo@ diData )
{
	// Gather data
	CBaseEntity@ attacker = diData.pAttacker;
	CBaseEntity@ inflictor = diData.pInflictor;
	CBasePlayer@ victim = cast< CBasePlayer@ >( g_EntityFuncs.Instance( diData.pVictim.pev ) );
	
	if ( victim.pev.FlagBitSet( FL_GODMODE ) )
	{
		// stop!
		diData.flDamage = 0.0;
		return HOOK_HANDLED;
	}
	
	if ( attacker is null )
		return HOOK_CONTINUE;
	
	if ( attacker.pev.classname == "monster_alien_slave" )
	{
		// give a knockback on their electric attacks
		if ( ( diData.bitsDamageType & DMG_SHOCK ) != 0 )
		{
			// Get the monsters aiming direction
			g_EngineFuncs.MakeVectors( attacker.pev.angles );
			
			// Calculate push velocity
			Vector vecVelocity = g_Engine.v_forward * 320;
			vecVelocity.z = 200; // Minimum speed required to lift the player off the ground
			
			// Push the player backwards!
			victim.pev.basevelocity = vecVelocity;
		}
	}
	
	victim.GetUserData( "flNextRegenTime" ) = g_Engine.time + 2.5;
	return HOOK_CONTINUE;
}

HookReturnCode PlayerPostThink( CBasePlayer@ pPlayer )
{
	// jumping is done on prethink, so check if we are off the ground
	if ( !pPlayer.pev.FlagBitSet( FL_ONGROUND ) && pPlayer.pev.button & IN_JUMP != 0 )
	{
		// don't pogo stick
		if ( int( pPlayer.GetUserData( "bDidJump" ) ) == 0 )
		{
			// extra jump boost
			if ( pPlayer.pev.rendercolor.y == 200 ) // G
				pPlayer.pev.velocity.z += 300;
			
			pPlayer.GetUserData( "bDidJump" ) = 1;
		}
	}
	
	// landing
	if ( pPlayer.pev.FlagBitSet( FL_ONGROUND ) && pPlayer.pev.health > 0 )
	{
		pPlayer.GetUserData( "bDidJump" ) = 0;
		
		// don't let the player stand above the barrel
		if ( g_hBoss.IsValid() )
		{
			CBaseEntity@ pBoss = g_hBoss.GetEntity();
			if ( pPlayer.pev.groundentity is pBoss.edict() )
				pPlayer.TakeDamage( pBoss.pev, pBoss.pev, 1.0, DMG_DROWN );
		}
	}
	
	return HOOK_CONTINUE;
}

HookReturnCode PlayerKilled( CBasePlayer@ pPlayer, CBaseEntity@ pAttacker, int iGib )
{
	// mark as dead
	pPlayer.pev.targetname = "dead_player";
	SURVIVAL::PlayerKilled( pPlayer, pAttacker, iGib );
	
	// no more living players?
	CBaseEntity@ pScan = g_EntityFuncs.RandomTargetname( "living_player" );
	if ( pScan is null && SURVIVAL::IsEnabled() )
	{
		// game over
		GameOver();
	}
	else
	{
		// attempt to respawn
		g_Scheduler.SetTimeout( "PlayerRespawn", g_EngineFuncs.CVarGetFloat( "mp_respawndelay" ), EHandle( pPlayer ) );
	}
	
	return HOOK_CONTINUE;
}

void PlayerRespawn( EHandle hPlayer )
{
	// survival enabled, abort
	if ( SURVIVAL::IsEnabled() )
		return;
	
	if ( hPlayer.IsValid() )
	{
		CBasePlayer@ pPlayer = cast< CBasePlayer@ >( hPlayer.GetEntity() );
		pPlayer.pev.targetname = "living_player";
		g_PlayerFuncs.RespawnPlayer( pPlayer, true, true ); // force-respawn
		
		pPlayer.GiveNamedItem( "weapon_ins2m2", SF_NORESPAWN );
		if ( g_iPhase == 5 )
			pPlayer.GiveNamedItem( "weapon_islave", SF_NORESPAWN );
		
		// give some spawn protection
		pPlayer.pev.flags |= FL_GODMODE;
		g_Scheduler.SetTimeout( "EndProtection", 1.5, EHandle( pPlayer ) );
	}
}
void EndProtection( EHandle hPlayer )
{
	if ( hPlayer.IsValid() )
	{
		CBasePlayer@ pPlayer = cast< CBasePlayer@ >( hPlayer.GetEntity() );
		pPlayer.pev.flags &= ~FL_GODMODE;
	}
}

void BossStart( CBaseEntity@ pActivator, CBaseEntity@ pCaller, USE_TYPE useType, float flValue )
{
	if ( g_hBoss.IsValid() )
	{
		// First attack is always crush
		CrushAttack();
	}
}

void GetEnemy()
{
	if ( g_hBoss.IsValid() )
	{
		CBaseEntity@ pBoss = g_hBoss.GetEntity();
		
		CBaseEntity@ pPlayer = g_EntityFuncs.RandomTargetname( "living_player" );
		if ( pPlayer !is null && pPlayer.IsAlive() && !pPlayer.pev.FlagBitSet( FL_GODMODE ) && !pPlayer.pev.FlagBitSet( FL_NOTARGET ) )
		{
			if ( pPlayer.pev.effects & EF_NODRAW != 0 )
			{
				// mark this monster
				CBaseEntity@ pMonster = FindEntityByOwner( null, "cnpc_islave", pPlayer.edict() );
				pMonster.pev.renderfx = kRenderFxGlowShell;
				pMonster.pev.rendercolor.x = 200; // R
				if ( pMonster.pev.rendercolor.y == 0 )
					pMonster.pev.rendercolor.y = 50; // G
				pMonster.pev.rendercolor.z = 50; // B
				pMonster.pev.renderamt = 1;
				@pBoss.pev.enemy = pMonster.edict();
			}
			else
			{
				// mark this player
				pPlayer.pev.renderfx = kRenderFxGlowShell;
				pPlayer.pev.rendercolor.x = 200; // R
				if ( pPlayer.pev.rendercolor.y == 0 )
					pPlayer.pev.rendercolor.y = 50; // G
				pPlayer.pev.rendercolor.z = 50; // B
				pPlayer.pev.renderamt = 1;
				@pBoss.pev.enemy = pPlayer.edict();
			}
		}
	}
}

void DoAttack( string szCallWhenMoveDone = "", float flTriggerDelay = 0.0f )
{
	if ( g_hBoss.IsValid() )
	{
		CBaseEntity@ pBoss = g_hBoss.GetEntity();
		
		CBaseEntity@ pEnemy = g_EntityFuncs.Instance( pBoss.pev.enemy );
		if ( pEnemy !is null && pEnemy.IsAlive() )
		{
			// attack
			pBoss.pev.takedamage = DAMAGE_YES;
			if ( pEnemy.pev.rendercolor.y == 50 ) // G
				pEnemy.pev.renderfx = kRenderFxNone;
			else
				pEnemy.pev.rendercolor.x = 50;
			g_EntityFuncs.SetOrigin( pBoss, Vector( pEnemy.pev.origin.x, pEnemy.pev.origin.y, 450 ) );
			
			Vector vecDest = Vector( pEnemy.pev.origin.x, pEnemy.pev.origin.y, g_vecFloor.z + ( pBoss.pev.size.z / 2 ) );
			LinearMove( pBoss, vecDest, BOSS_CRUSH_SPEED, szCallWhenMoveDone, flTriggerDelay );
		}
		else
		{
			// No target, try again
			GetEnemy();
			g_Scheduler.SetTimeout( "DoAttack", 1.0, szCallWhenMoveDone, flTriggerDelay );
		}
	}
}

void CrushAttack()
{
	if ( g_hBoss.IsValid() )
	{
		GetEnemy();
		g_Scheduler.SetTimeout( "DoAttack", 1.0, "HideBarrel", 1.0f );
	}
}

void MinionAttack()
{
	if ( g_hBoss.IsValid() )
	{
		GetEnemy();
		g_Scheduler.SetTimeout( "DoAttack", 1.0, "MinionSpawn", 0.0f );
	}
}
void MinionSpawn()
{
	if ( g_hBoss.IsValid() )
	{
		CBaseEntity@ pBoss = g_hBoss.GetEntity();
		CBaseEntity@ pEnemy = g_EntityFuncs.Instance( pBoss.pev.enemy );
		
		int iPlayers = g_PlayerFuncs.GetNumPlayers();
		
		// Face the enemy direction if possible, otherwise just pick a random angle
		Vector vecAngles = Vector( 0, Math.RandomLong( -180, 180 ), 0 );
		if ( pEnemy !is null && pEnemy.IsAlive() )
			vecAngles.y = Math.VecToYaw( pEnemy.pev.origin - pBoss.pev.origin );
		
		// Spawn 3 monsters in void room, then move it inside arena
		CBaseMonster@ pMonster;
		Vector vecOrigin = g_vecZero;
		
		// If many players, spawn a 4th
		uint NUM_MONSTERS = 3 + ( iPlayers > 4 ? 1 : 0 );
		array< bool > bSpawnUsed( 4 );
		
		for ( uint uiMonster = 0; uiMonster < NUM_MONSTERS; uiMonster++ )
		{
			@pMonster = cast< CBaseMonster@ >( g_EntityFuncs.Create( "monster_alien_slave", g_vecVoid, vecAngles, false ) );
			if ( iPlayers > 1 )
			{
				// every extra player increases monster health by +12%
				pMonster.pev.health += pMonster.pev.health * ( ( float( iPlayers ) - 1.0 ) * 0.12 );
				pMonster.pev.max_health = pMonster.pev.health;
			}
			pMonster.KeyValue( "is_not_revivable", "1" );
			pMonster.pev.targetname = "giegue_minion_npc";
			pMonster.StartMonster(); // think now, we want the entity to drop to ground and be ready
			pMonster.pev.nextthink = g_Engine.time;
			
			while ( vecOrigin == g_vecZero )
			{
				switch ( Math.RandomLong( 0, 3 ) )
				{
					case 0:
					{
						if ( !bSpawnUsed[ 0 ] )
						{
							vecOrigin.x = pBoss.pev.origin.x + ( pBoss.pev.size.x * -1 );
							vecOrigin.y = pBoss.pev.origin.y + pBoss.pev.size.y;
							pMonster.pev.vuser1 = Vector( -135, 135, 270 );
							bSpawnUsed[ 0 ] = true;
							break;
						}
					}
					case 1:
					{
						if ( !bSpawnUsed[ 1 ] )
						{
							vecOrigin.x = pBoss.pev.origin.x + pBoss.pev.size.x;
							vecOrigin.y = pBoss.pev.origin.y + ( pBoss.pev.size.y * -1 );
							pMonster.pev.vuser1 = Vector( 135, -135, 270 );
							bSpawnUsed[ 1 ] = true;
							break;
						}
					}
					case 2:
					{
						if ( !bSpawnUsed[ 2 ] )
						{
							vecOrigin.x = pBoss.pev.origin.x + ( pBoss.pev.size.x * -1 );
							vecOrigin.y = pBoss.pev.origin.y + ( pBoss.pev.size.y * -1 );
							pMonster.pev.vuser1 = Vector( -135, -135, 270 );
							bSpawnUsed[ 2 ] = true;
							break;
						}
					}
					case 3:
					{
						if ( !bSpawnUsed[ 3 ] )
						{
							vecOrigin.x = pBoss.pev.origin.x + pBoss.pev.size.x;
							vecOrigin.y = pBoss.pev.origin.y + pBoss.pev.size.y;
							pMonster.pev.vuser1 = Vector( 135, 135, 270 );
							bSpawnUsed[ 3 ] = true;
							break;
						}
					}
				}
			}
			
			vecOrigin.z = g_vecFloor.z + ( pBoss.pev.size.z / 2 );
			g_EntityFuncs.SetOrigin( pMonster, vecOrigin );
			
			pMonster.pev.velocity = pMonster.pev.vuser1;
			vecOrigin = g_vecZero;
		}
		
		g_SoundSystem.EmitSoundDyn( pBoss.edict(), CHAN_WEAPON, "player/pl_long_jump.wav", VOL_NORM, ATTN_NORM, 0, PITCH_NORM );
		g_Scheduler.SetTimeout( "HideBarrel", 2.0 );
	}
}

void SingleGLAttack()
{
	if ( g_hBoss.IsValid() )
	{
		GetEnemy();
		g_Scheduler.SetTimeout( "DoAttack", g_iPhase == 6 ? 0.8 : 1.0, "SingleGLCluster", 0.0f );
	}
}
void SingleGLCluster()
{
	if ( g_hBoss.IsValid() )
	{
		CBaseEntity@ pBoss = g_hBoss.GetEntity();
		CGrenade@ pGrenade = null;
		
		for ( uint uiGrenade = 0; uiGrenade < 4; uiGrenade++ )
		{
			Vector vecOrigin = g_vecZero;
			vecOrigin.z = g_vecFloor.z + ( pBoss.pev.size.z / 2 );
			switch ( uiGrenade )
			{
				case 0:
				{
					vecOrigin.x = pBoss.pev.origin.x + ( pBoss.pev.size.x * -1 );
					vecOrigin.y = pBoss.pev.origin.y + pBoss.pev.size.y;
					
					@pGrenade = g_EntityFuncs.ShootContact( pBoss.pev, vecOrigin, Vector( -130, 130, 270 ) );
					break;
				}
				case 1:
				{
					vecOrigin.x = pBoss.pev.origin.x + pBoss.pev.size.x;
					vecOrigin.y = pBoss.pev.origin.y + ( pBoss.pev.size.y * -1 );
					
					@pGrenade = g_EntityFuncs.ShootContact( pBoss.pev, vecOrigin, Vector( 130, -130, 270 ) );
					break;
				}
				case 2:
				{
					vecOrigin.x = pBoss.pev.origin.x + ( pBoss.pev.size.x * -1 );
					vecOrigin.y = pBoss.pev.origin.y + ( pBoss.pev.size.y * -1 );
					
					@pGrenade = g_EntityFuncs.ShootContact( pBoss.pev, vecOrigin, Vector( -130, -130, 270 ) );
					break;
				}
				case 3:
				{
					vecOrigin.x = pBoss.pev.origin.x + pBoss.pev.size.x;
					vecOrigin.y = pBoss.pev.origin.y + pBoss.pev.size.y;
					
					@pGrenade = g_EntityFuncs.ShootContact( pBoss.pev, vecOrigin, Vector( 130, 130, 270 ) );
					break;
				}
			}
			pGrenade.pev.renderfx = kRenderFxGlowShell;
			pGrenade.pev.rendercolor = Vector( 250, 230, 50 );
			pGrenade.pev.renderamt = 255; // T H I C C
			pGrenade.pev.dmg = 56; // default damage has a lot of radius
		}
		
		switch ( Math.RandomLong( 1, 2 ) )
		{
			case 1: g_SoundSystem.EmitSoundDyn( pBoss.edict(), CHAN_WEAPON, "weapons/glauncher.wav", VOL_NORM, ATTN_NORM, 0, PITCH_NORM ); break;
			case 2: g_SoundSystem.EmitSoundDyn( pBoss.edict(), CHAN_WEAPON, "weapons/glauncher2.wav", VOL_NORM, ATTN_NORM, 0, PITCH_NORM ); break;
		}
		g_Scheduler.SetTimeout( "HideBarrel", g_iPhase == 6 ? 0.75 : 1.5 );
	}
}

void BounceAttack()
{
	if ( g_hBoss.IsValid() )
	{
		CBaseEntity@ pBoss = g_hBoss.GetEntity();
		
		CBaseEntity@ pEnemy = g_EntityFuncs.Instance( pBoss.pev.enemy );
		if ( pEnemy !is null && pEnemy.IsAlive() )
		{
			// attack
			pBoss.pev.takedamage = DAMAGE_YES;
			if ( pEnemy.pev.rendercolor.y == 50 ) // G
				pEnemy.pev.renderfx = kRenderFxNone;
			else
				pEnemy.pev.rendercolor.x = 50;
			g_EntityFuncs.SetOrigin( pBoss, Vector( pEnemy.pev.origin.x, pEnemy.pev.origin.y, 450 ) );
			
			pBoss.pev.movetype = MOVETYPE_PUSHSTEP;
			INS2FLAMEPROJECTILE::SOLID_TYPE = SOLID_NOT; // prevent the flame projectiles from stopping the barrel upon collision
			ScanBossBounce();
			ScanFlameHit();
		}
		else
		{
			// No target, try again
			GetEnemy();
			g_Scheduler.SetTimeout( "BounceAttack", 0.9 );
		}
	}
}
void ScanBossBounce()
{
	if ( g_hBoss.IsValid() )
	{
		CBaseEntity@ pBoss = g_hBoss.GetEntity();
		
		if ( pBoss.pev.FlagBitSet( FL_ONGROUND ) )
		{
			// select a random direction to bounce to
			Vector vecDest = g_vecZero;
			vecDest.x = Math.RandomFloat( g_vecFloor.x - 432, g_vecFloor.x + 432 );
			vecDest.y = Math.RandomFloat( g_vecFloor.y - 432, g_vecFloor.y + 432 );
			
			// setup travel speed
			Vector vecVelocity = ( vecDest - pBoss.pev.origin ).Normalize() * ( BOSS_CRUSH_SPEED / 2.0 );
			vecVelocity.z = 460; // 465 // 480
			pBoss.pev.velocity = vecVelocity;
			
			// spawn a spore
			CBaseEntity@ pSpore = g_EntityFuncs.Create( "sporegrenade", pBoss.pev.origin + Vector( 0, 0, pBoss.pev.size.z ), g_vecZero, false );
			pSpore.pev.solid = SOLID_NOT; // do not collide with the barrel
			pSpore.pev.scale = 5.55;
			pSpore.pev.velocity = Vector( Math.RandomFloat( -270, 270 ), Math.RandomFloat( -270, 270 ), 270 );
			
			// sound effect, and keep track of spawned spores
			g_SoundSystem.EmitSoundDyn( pBoss.edict(), CHAN_WEAPON, "misc/bbimpact.wav", VOL_NORM, ATTN_NONE, 0, PITCH_NORM );
			g_iSpores++;
			
			// remove it manually
			g_Scheduler.SetTimeout( "DeleteSpore", 5.10, EHandle( pSpore ) );
		}
		
		if ( pBoss.pev.health > PHASE_HP_THRESHOLD )
			g_Scheduler.SetTimeout( "ScanBossBounce", 0.05 );
	}
}
void ScanFlameHit()
{
	if ( g_hBoss.IsValid() )
	{
		CBaseEntity@ pBoss = g_hBoss.GetEntity();
		
		if ( g_bIntermission )
			return;
		
		CBaseEntity@ pFlame = null;
		while ( ( @pFlame = g_EntityFuncs.FindEntityByClassname( pFlame, "proj_ins2flame" ) ) !is null )
		{
			if ( int( pFlame.GetUserData( "touch" ) ) == 0 && pFlame.Intersects( pBoss ) )
			{
				// assume touch
				pFlame.Touch( pBoss );
				pFlame.GetUserData( "touch" ) = 1;
			}
		}
		
		g_Scheduler.SetTimeout( "ScanFlameHit", 0.05 );
	}
}

void DeleteSpore( EHandle hSpore )
{
	if ( hSpore.IsValid() )
	{
		CBaseEntity@ pSpore = hSpore.GetEntity();
		g_EntityFuncs.Remove( pSpore );
	}
}

void CountAttack()
{
	if ( g_hBoss.IsValid() )
	{
		CBaseEntity@ pBoss = g_hBoss.GetEntity();
		
		g_SoundSystem.EmitSoundDyn( pBoss.edict(), CHAN_WEAPON, "ragemap2020/ClockTick.wav", VOL_NORM, ATTN_NONE, SND_FORCE_LOOP, PITCH_NORM );
		
		TIMER::Show( 5.0f );
		g_Scheduler.SetTimeout( "CountCheck", 5.0f );
	}
}
void CountCheck()
{
	if ( g_hBoss.IsValid() )
	{
		CBaseEntity@ pBoss = g_hBoss.GetEntity();
		
		g_SoundSystem.StopSound( pBoss.edict(), CHAN_WEAPON, "ragemap2020/ClockTick.wav" );
		
		// trigger the game_zone_player
		g_EntityFuncs.FireTargets( "giegue_math_left", null, null, USE_TOGGLE, 0.0f, 0.0f );
		g_EntityFuncs.FireTargets( "giegue_math_middle", null, null, USE_TOGGLE, 0.0f, 0.0f );
		g_EntityFuncs.FireTargets( "giegue_math_right", null, null, USE_TOGGLE, 0.0f, 0.0f );
		
		g_EntityFuncs.FireTargets( "giegue_math_attack_mm", null, null, USE_TOGGLE, 0.0f, 0.01f );
	}
}
void CountFail( CBaseEntity@ pActivator, CBaseEntity@ pCaller, USE_TYPE useType, float flValue )
{
	if ( g_hBoss.IsValid() )
	{
		CBaseEntity@ pBoss = g_hBoss.GetEntity();
		if ( pActivator.IsPlayer() && pActivator.IsAlive() )
		{
			pActivator.Killed( pBoss.pev, GIB_ALWAYS );
			
			// one player dying is enough to call this entity
			g_EntityFuncs.FireTargets( "giegue_math_krelay", null, null, USE_TOGGLE, 0.0f, 0.0f );
		}
	}
}

void DoFloorAttack()
{
	if ( g_hBoss.IsValid() )
	{
		CBaseEntity@ pBoss = g_hBoss.GetEntity();
		
		CBaseEntity@ pEnemy = g_EntityFuncs.Instance( pBoss.pev.enemy );
		if ( pEnemy !is null && pEnemy.IsAlive() )
		{
			// attack
			pBoss.pev.takedamage = DAMAGE_YES;
			if ( pEnemy.pev.rendercolor.y == 50 ) // G
				pEnemy.pev.renderfx = kRenderFxNone;
			else
				pEnemy.pev.rendercolor.x = 50;
			g_EntityFuncs.SetOrigin( pBoss, Vector( pEnemy.pev.origin.x, pEnemy.pev.origin.y, 450 ) );
			
			Vector vecDest = Vector( pEnemy.pev.origin.x, pEnemy.pev.origin.y, g_vecFloor.z + ( pBoss.pev.size.z / 2 ) );
			LinearMove( pBoss, vecDest, BOSS_CRUSH_SPEED, "RandomFloorAttack", 0.5 );
		}
		else
		{
			// No target, try again
			GetEnemy();
			g_Scheduler.SetTimeout( "DoFloorAttack", 0.8 );
		}
	}
}
void RandomFloorAttack()
{
	if ( g_bIntermission || g_iPhase != 4 )
		return;
	
	if ( g_hBoss.IsValid() )
	{
		CBaseEntity@ pBoss = g_hBoss.GetEntity();
		
		if ( pBoss.pev.velocity == g_vecZero )
			RandomFloorMove();
		
		int iPlayers = g_PlayerFuncs.GetNumPlayers();
		
		switch ( Math.RandomLong( 1, 3 ) )
		{
			case 1: // Water Darts
			{
				// If many players, spawn 12 darts
				uint NUM_DARTS = 8 + ( iPlayers > 4 ? 4 : 0 );
				
				Vector vecAngles = g_vecZero;
				
				Vector vecOrigin = pBoss.pev.origin;
				vecOrigin.z = g_vecFloor.z + ( pBoss.pev.size.z / 2 );
				
				for ( uint uiDart = 0; uiDart < NUM_DARTS; uiDart++ )
				{
					// Calculate forward velocity
					vecAngles.x = Math.RandomFloat( -45, 45 ); // Tilt the darts a bit
					g_EngineFuncs.MakeVectors( vecAngles );
					
					Vector vecForward = g_Engine.v_forward;
					Vector vecVelocity = vecForward * BOLT_SPEED;
					
					// Create dart
					CBaseEntity@ pDart = g_EntityFuncs.Create( "bounce_dart", vecOrigin + vecForward * ( pBoss.pev.size.x / 2 ), vecAngles, false, pBoss.edict() );
					pDart.pev.targetname = "giegue_dart_projectile";
					
					// Fly away!
					pDart.pev.velocity = vecVelocity;
					
					vecAngles.y += NUM_DARTS == 12 ? 30 : 45;
				}
				
				g_SoundSystem.EmitSoundDyn( pBoss.edict(), CHAN_WEAPON, "weapons/xbow_fire1.wav", VOL_NORM, ATTN_NORM, 0, 93 + Math.RandomLong( 0, 0xF ) );
				g_Scheduler.SetTimeout( "RandomFloorAttack", Math.RandomFloat( 10.6, 11.0 ) );
				break;
			}
			case 2: // Mortar
			{
				// If many players, spawn a 5th
				int NUM_PROJECTILES = 4 + ( iPlayers > 4 ? 1 : 0 );
				
				for ( int iMortar = 0; iMortar < NUM_PROJECTILES; iMortar++ )
				{
					Vector vecOrigin = pBoss.pev.origin + Vector( iMortar, iMortar, pBoss.pev.size.z / 2 );
					
					// randomize velocity a bit
					Vector vecVelocity = Vector( Math.RandomFloat( -140, 140 ), Math.RandomFloat( -140, 140 ), 200 + ( 130 * iMortar ) );
					
					CBMortar::Shoot( pBoss.edict(), vecOrigin, vecVelocity );
				}
				
				g_Scheduler.SetTimeout( "RandomFloorAttack", Math.RandomFloat( 0.7, 2.8 ) );
				break;
			}
			case 3: // Minions
			{
				// Pick a random angle
				Vector vecAngles = Vector( 0, Math.RandomLong( -180, 180 ), 0 );
				
				// Spawn 3 monsters in void room, then move it inside arena
				CBaseMonster@ pMonster;
				Vector vecOrigin = g_vecZero;
				
				array< bool > bSpawnUsed( 4 );
				
				for ( uint uiMonster = 0; uiMonster < 3; uiMonster++ )
				{
					@pMonster = cast< CBaseMonster@ >( g_EntityFuncs.Create( "monster_alien_grunt", g_vecVoid, vecAngles, false ) );
					pMonster.KeyValue( "is_not_revivable", "1" );
					pMonster.pev.targetname = "giegue_minion_npc";
					pMonster.pev.weapons = 8; // melee-only
					pMonster.pev.body = 1; // no hive-hand
					pMonster.StartMonster(); // think now, we want the entity to drop to ground and be ready
					pMonster.pev.nextthink = g_Engine.time;
					
					while ( vecOrigin == g_vecZero )
					{
						switch ( Math.RandomLong( 0, 3 ) )
						{
							case 0:
							{
								if ( !bSpawnUsed[ 0 ] )
								{
									vecOrigin.x = pBoss.pev.origin.x + ( pBoss.pev.size.x * -1 );
									vecOrigin.y = pBoss.pev.origin.y + pBoss.pev.size.y;
									pMonster.pev.vuser1 = Vector( -135, 135, 270 );
									bSpawnUsed[ 0 ] = true;
									break;
								}
							}
							case 1:
							{
								if ( !bSpawnUsed[ 1 ] )
								{
									vecOrigin.x = pBoss.pev.origin.x + pBoss.pev.size.x;
									vecOrigin.y = pBoss.pev.origin.y + ( pBoss.pev.size.y * -1 );
									pMonster.pev.vuser1 = Vector( 135, -135, 270 );
									bSpawnUsed[ 1 ] = true;
									break;
								}
							}
							case 2:
							{
								if ( !bSpawnUsed[ 2 ] )
								{
									vecOrigin.x = pBoss.pev.origin.x + ( pBoss.pev.size.x * -1 );
									vecOrigin.y = pBoss.pev.origin.y + ( pBoss.pev.size.y * -1 );
									pMonster.pev.vuser1 = Vector( -135, -135, 270 );
									bSpawnUsed[ 2 ] = true;
									break;
								}
							}
							case 3:
							{
								if ( !bSpawnUsed[ 3 ] )
								{
									vecOrigin.x = pBoss.pev.origin.x + pBoss.pev.size.x;
									vecOrigin.y = pBoss.pev.origin.y + pBoss.pev.size.y;
									pMonster.pev.vuser1 = Vector( 135, 135, 270 );
									bSpawnUsed[ 3 ] = true;
									break;
								}
							}
						}
					}
					
					vecOrigin.z = g_vecFloor.z + ( pBoss.pev.size.z / 2 );
					g_EntityFuncs.SetOrigin( pMonster, vecOrigin );
					
					pMonster.pev.velocity = pMonster.pev.vuser1;
					vecOrigin = g_vecZero;
				}
				
				g_SoundSystem.EmitSoundDyn( pBoss.edict(), CHAN_WEAPON, "player/pl_long_jump.wav", VOL_NORM, ATTN_NORM, 0, PITCH_NORM );
				g_Scheduler.SetTimeout( "RandomFloorAttack", Math.RandomFloat( 3.5, 7.0 ) );
				break;
			}
		}
	}
}
void RandomFloorMove()
{
	if ( g_bIntermission || g_iPhase != 4 )
		return;
	
	if ( g_hBoss.IsValid() )
	{
		CBaseEntity@ pBoss = g_hBoss.GetEntity();
		
		// select a random direction to travel to
		Vector vecDest = g_vecZero;
		vecDest.x = Math.RandomFloat( g_vecFloor.x - 448, g_vecFloor.x + 448 ); // go a little off the road to be able to push players off the edge
		vecDest.y = Math.RandomFloat( g_vecFloor.y - 448, g_vecFloor.y + 448 );
		vecDest.z = pBoss.pev.origin.z;
		
		int SPEED = Math.RandomLong( 1, 3 ) == 1 ? BOSS_CRUSH_SPEED : BOSS_CRUSH_SPEED / 2;
		
		// go there
		LinearMove( pBoss, vecDest, SPEED, "RandomFloorMove", 0.0 );
	}
}

void DoTransformAttack()
{
	if ( g_hBoss.IsValid() )
	{
		CBaseEntity@ pBoss = g_hBoss.GetEntity();
		
		CBaseEntity@ pEnemy = g_EntityFuncs.Instance( pBoss.pev.enemy );
		if ( pEnemy !is null && pEnemy.IsAlive() )
		{
			// only move X/Y
			if ( pEnemy.pev.rendercolor.y == 50 ) // G
				pEnemy.pev.renderfx = kRenderFxNone;
			else
				pEnemy.pev.rendercolor.x = 50;
			
			Vector vecDest = Vector( pEnemy.pev.origin.x, pEnemy.pev.origin.y, pBoss.pev.origin.z );
			LinearMove( pBoss, vecDest, BOSS_CRUSH_SPEED, "TransformMinionAttack", 0.0 );
		}
		else
		{
			// No target, try again
			GetEnemy();
			g_Scheduler.SetTimeout( "DoTransformAttack", 0.7 );
		}
	}
}
void TransformMinionAttack()
{
	if ( g_hBoss.IsValid() )
	{
		CBaseEntity@ pBoss = g_hBoss.GetEntity();
		CBaseEntity@ pEnemy = g_EntityFuncs.Instance( pBoss.pev.enemy );
		
		int iPlayers = g_PlayerFuncs.GetNumPlayers();
		
		// Face the enemy direction if possible, otherwise just pick a random angle
		Vector vecAngles = Vector( 0, Math.RandomLong( -180, 180 ), 0 );
		if ( pEnemy !is null && pEnemy.IsAlive() )
			vecAngles.y = Math.VecToYaw( pEnemy.pev.origin - pBoss.pev.origin );
		
		CBaseMonster@ pMonster;
		Vector vecOrigin = g_vecZero;
		
		// spawn 4 monsters
		for ( uint uiMonster = 0; uiMonster < 4; uiMonster++ )
		{
			@pMonster = cast< CBaseMonster@ >( g_EntityFuncs.Create( "monster_zombie", g_vecVoid, vecAngles, false ) );
			pMonster.pev.health = pMonster.pev.health = 25; // one zap attack should be able to kill it
			pMonster.KeyValue( "is_not_revivable", "1" );
			pMonster.pev.targetname = "giegue_minion_npc";
			pMonster.StartMonster(); // think now, we want the entity to drop to ground and be ready
			pMonster.pev.nextthink = g_Engine.time;
			
			vecOrigin.z = pBoss.pev.origin.z + ( pBoss.pev.size.z / 2 );
			switch ( uiMonster )
			{
				case 0:
				{
					vecOrigin.x = pBoss.pev.origin.x + ( pBoss.pev.size.x * -1 );
					vecOrigin.y = pBoss.pev.origin.y + pBoss.pev.size.y;
					
					g_EntityFuncs.SetOrigin( pMonster, vecOrigin );
					
					pMonster.pev.velocity = Vector( -65, 65, 135 );
					break;
				}
				case 1:
				{
					vecOrigin.x = pBoss.pev.origin.x + pBoss.pev.size.x;
					vecOrigin.y = pBoss.pev.origin.y + ( pBoss.pev.size.y * -1 );
					
					g_EntityFuncs.SetOrigin( pMonster, vecOrigin );
					
					pMonster.pev.velocity = Vector( 65, -65, 135 );
					break;
				}
				case 2:
				{
					vecOrigin.x = pBoss.pev.origin.x + ( pBoss.pev.size.x * -1 );
					vecOrigin.y = pBoss.pev.origin.y + ( pBoss.pev.size.y * -1 );
					
					g_EntityFuncs.SetOrigin( pMonster, vecOrigin );
					
					pMonster.pev.velocity = Vector( -65, -65, 135 );
					break;
				}
				case 3:
				{
					vecOrigin.x = pBoss.pev.origin.x + pBoss.pev.size.x;
					vecOrigin.y = pBoss.pev.origin.y + pBoss.pev.size.y;
					
					g_EntityFuncs.SetOrigin( pMonster, vecOrigin );
					
					pMonster.pev.velocity = Vector( 65, 65, 135 );
					break;
				}
			}
		}
		
		g_SoundSystem.EmitSoundDyn( pBoss.edict(), CHAN_WEAPON, "player/pl_long_jump.wav", VOL_NORM, ATTN_NORM, 0, PITCH_NORM );
		
		// how many times did we do this?
		int iSpawns = int( pBoss.GetUserData( "spawns" ) );
		
		// repeat this many times
		int iMaxSpawns = 4 + ( iPlayers > 4 ? 1 : 0 );
		
		iSpawns++;
		pBoss.GetUserData( "spawns" ) = iSpawns;
		
		if ( iSpawns < iMaxSpawns )
		{
			GetEnemy();
			g_Scheduler.SetTimeout( "DoTransformAttack", 1.25 );
		}
		else
		{
			g_Scheduler.SetTimeout( "TransformCrushAttackPre", 2.50 );
			g_Scheduler.SetTimeout( "ScanZombieCrush", 0.1 );
		}
	}
}

void TransformCrushAttackPre()
{
	if ( g_hBoss.IsValid() )
	{
		GetEnemy();
		g_Scheduler.SetTimeout( "TransformCrushAttack", 1.0 );
	}
}
void TransformCrushAttack()
{
	if ( g_hBoss.IsValid() )
	{
		CBaseEntity@ pBoss = g_hBoss.GetEntity();
		
		CBaseEntity@ pEnemy = g_EntityFuncs.Instance( pBoss.pev.enemy );
		if ( pEnemy !is null && pEnemy.IsAlive() )
		{
			// only move X/Y
			if ( pEnemy.pev.rendercolor.y == 50 ) // G
				pEnemy.pev.renderfx = kRenderFxNone;
			else
				pEnemy.pev.rendercolor.x = 50;
			
			Vector vecDest = Vector( pEnemy.pev.origin.x, pEnemy.pev.origin.y, pBoss.pev.origin.z );
			LinearMove( pBoss, vecDest, BOSS_CRUSH_SPEED, "TransformCrushAttackPost", 0.0 );
		}
		else
		{
			// No target, try again
			GetEnemy();
			g_Scheduler.SetTimeout( "TransformCrushAttack", 0.7 );
		}
	}
}
void TransformCrushAttackPost()
{
	if ( g_hBoss.IsValid() )
	{
		CBaseEntity@ pBoss = g_hBoss.GetEntity();
		
		CBaseEntity@ pEnemy = g_EntityFuncs.Instance( pBoss.pev.enemy );
		if ( pEnemy !is null && pEnemy.IsAlive() )
		{
			// don't do this forever
			int iMinions = 0;
			CBaseEntity@ pMinion = null;
			while ( ( @pMinion = g_EntityFuncs.FindEntityByTargetname( pMinion, "giegue_minion_npc" ) ) !is null )
			{
				if ( pMinion.IsAlive() )
					iMinions++;
			}
			
			// only move Z
			Vector vecDest = Vector( pBoss.pev.origin.x, pBoss.pev.origin.y, g_vecFloor.z + ( pBoss.pev.size.z / 2 ) );
			LinearMove( pBoss, vecDest, BOSS_CRUSH_SPEED, iMinions == 0 ? "HideBarrel" : "LiftBarrel", 1.25 );
		}
		else
		{
			// No target, try again
			GetEnemy();
			g_Scheduler.SetTimeout( "TransformCrushAttack", 0.6 );
		}
	}
}
void ScanZombieCrush()
{
	if ( g_hBoss.IsValid() && g_hHurt.IsValid() )
	{
		CBaseEntity@ pBoss = g_hBoss.GetEntity();
		CBaseEntity@ pHurt = g_hHurt.GetEntity();
		
		if ( pBoss.pev.origin.z == 450 )
			return;
		
		CBaseEntity@ pMonster = null;
		while ( ( @pMonster = g_EntityFuncs.FindEntityByClassname( pMonster, "monster_zombie" ) ) !is null )
		{
			if ( pHurt.Intersects( pMonster ) )
			{
				// assume touch, trigger is player only, so manually kill it
				pMonster.TakeDamage( pBoss.pev, pHurt.pev, 512.0, DMG_CRUSH );
			}
		}
		
		g_Scheduler.SetTimeout( "ScanZombieCrush", 0.1 );
	}
}

void TransformBoidAttack( int iSpawns = 0 )
{
	CBaseEntity@ pMonster;
	CBaseEntity@ pHurt;
	
	Vector vecOrigin = g_vecZero;
	vecOrigin.z = g_vecFloor.z + 32;
	
	// spawn 4 monsters
	for ( uint uiMonster = 0; uiMonster < 4; uiMonster++ )
	{
		@pMonster = g_EntityFuncs.Create( "monster_flyer", g_vecVoid, g_vecZero, false );
		pMonster.pev.solid = SOLID_NOT;
		pMonster.pev.movetype = MOVETYPE_NOCLIP;
		
		@pHurt = g_EntityFuncs.Create( "trigger_hurt", g_vecVoid, g_vecZero, true );
		g_EntityFuncs.SetModel( pHurt, g_EntityFuncs.FindEntityByTargetname( null, "giegue_boid_hurt_template" ).pev.model );
		pHurt.KeyValue( "damagetype", "4" ); // DMG_SLASH
		pHurt.pev.dmg = 256;
		g_EntityFuncs.DispatchSpawn( pHurt.edict() );
		pHurt.pev.movetype = MOVETYPE_FOLLOW;
		@pHurt.pev.aiment = pMonster.edict();
		
		g_Scheduler.SetTimeout( "DeleteBoid", 10.0, EHandle( pMonster ), EHandle( pHurt ) );
		
		switch ( uiMonster )
		{
			case 0:
			{
				vecOrigin.x = g_vecFloor.x - 544;
				vecOrigin.y = g_vecFloor.y + Math.RandomFloat( -448, 448 );
				
				g_EntityFuncs.SetOrigin( pMonster, vecOrigin );
				
				pMonster.pev.angles.y = 0;
				pMonster.pev.velocity.x = Math.RandomFloat( 125, 210 );
				break;
			}
			case 1:
			{
				vecOrigin.x = g_vecFloor.x + 544;
				vecOrigin.y = g_vecFloor.y + Math.RandomFloat( -448, 448 );
				
				g_EntityFuncs.SetOrigin( pMonster, vecOrigin );
				
				pMonster.pev.angles.y = 180;
				pMonster.pev.velocity.x = Math.RandomFloat( -210, -125 );
				break;
			}
			case 2:
			{
				vecOrigin.x = g_vecFloor.x + Math.RandomFloat( -448, 448 );
				vecOrigin.y = g_vecFloor.y - 544;
				
				g_EntityFuncs.SetOrigin( pMonster, vecOrigin );
				
				pMonster.pev.angles.y = 90;
				pMonster.pev.velocity.y = Math.RandomFloat( 125, 210 );
				break;
			}
			case 3:
			{
				vecOrigin.x = g_vecFloor.x + Math.RandomFloat( -448, 448 );
				vecOrigin.y = g_vecFloor.y + 544;
				
				g_EntityFuncs.SetOrigin( pMonster, vecOrigin );
				
				pMonster.pev.angles.y = 270;
				pMonster.pev.velocity.y = Math.RandomFloat( -210, -125 );
				break;
			}
		}
		pMonster.pev.nextthink = -1;
		pMonster.pev.frame = pMonster.pev.framerate = pMonster.pev.animtime = 1.0;
	}
	
	iSpawns++;
	if ( iSpawns < 9 )
		g_Scheduler.SetTimeout( "TransformBoidAttack", 1.30, iSpawns );
	else
		g_Scheduler.SetTimeout( "PreludePhase6", 10.0 );
}
void DeleteBoid( EHandle hMonster, EHandle hHurt )
{
	if ( hHurt.IsValid() )
	{
		CBaseEntity@ pHurt = hHurt.GetEntity();
		g_EntityFuncs.Remove( pHurt );
	}
	
	if ( hMonster.IsValid() )
	{
		CBaseEntity@ pMonster = hMonster.GetEntity();
		g_EntityFuncs.Remove( pMonster );
	}
}

void MultiGLAttack()
{
	if ( g_hBoss.IsValid() )
	{
		GetEnemy();
		g_Scheduler.SetTimeout( "DoAttack", 0.8, "MultiGLCluster", 0.0f );
	}
}
void MultiGLCluster()
{
	if ( g_hBoss.IsValid() )
	{
		CBaseEntity@ pBoss = g_hBoss.GetEntity();
		CGrenade@ pGrenade = null;
		
		// how many times did we do this?
		int iSpawns = int( pBoss.GetUserData( "grenades" ) );
		
		for ( uint uiGrenade = 0; uiGrenade < 4; uiGrenade++ )
		{
			Vector vecOrigin = pBoss.pev.origin;
			vecOrigin.z = g_vecFloor.z + ( pBoss.pev.size.z / 2 );
			
			Vector vecVelocity = g_vecZero;
			vecVelocity.z = 270;
			
			switch ( uiGrenade )
			{
				case 0:
				{
					vecVelocity.x = Math.RandomFloat( -65, -135 );
					vecVelocity.y = Math.RandomFloat( 65, 135 );
					
					vecVelocity = Math.RotateVector( vecVelocity, Vector( 0, 60 * iSpawns, 0 ), g_vecZero );
					Math.MakeVectors( Vector( 0, Math.VecToYaw( vecVelocity ), 0 ) );
					
					vecOrigin = vecOrigin + g_Engine.v_forward * 64;
					break;
				}
				case 1:
				{
					vecVelocity.x = Math.RandomFloat( 65, 135 );
					vecVelocity.y = Math.RandomFloat( -65, -135 );
					
					vecVelocity = Math.RotateVector( vecVelocity, Vector( 0, 60 * ( iSpawns + 1 ), 0 ), g_vecZero );
					Math.MakeVectors( Vector( 0, Math.VecToYaw( vecVelocity ), 0 ) );
					
					vecOrigin = vecOrigin + g_Engine.v_forward * 64;
					break;
				}
				case 2:
				{
					vecVelocity.x = Math.RandomFloat( -65, -135 );
					vecVelocity.y = Math.RandomFloat( -65, -135 );
					
					vecVelocity = Math.RotateVector( vecVelocity, Vector( 0, 60 * ( iSpawns + 2 ), 0 ), g_vecZero );
					Math.MakeVectors( Vector( 0, Math.VecToYaw( vecVelocity ), 0 ) );
					
					vecOrigin = vecOrigin + g_Engine.v_forward * 64;
					break;
				}
				case 3:
				{
					vecVelocity.x = Math.RandomFloat( 65, 135 );
					vecVelocity.y = Math.RandomFloat( 65, 135 );
					
					vecVelocity = Math.RotateVector( vecVelocity, Vector( 0, 60 * ( iSpawns + 3 ), 0 ), g_vecZero );
					Math.MakeVectors( Vector( 0, Math.VecToYaw( vecVelocity ), 0 ) );
					
					vecOrigin = vecOrigin + g_Engine.v_forward * 64;
					break;
				}
			}
			@pGrenade = g_EntityFuncs.ShootContact( pBoss.pev, vecOrigin, vecVelocity );
			pGrenade.pev.renderfx = kRenderFxGlowShell;
			pGrenade.pev.rendercolor = Vector( 250, 230, 50 );
			pGrenade.pev.renderamt = 255; // T H I C C
			pGrenade.pev.dmg = 52; // too many grenades, do even lower damage
		}
		
		switch ( Math.RandomLong( 1, 2 ) )
		{
			case 1: g_SoundSystem.EmitSoundDyn( pBoss.edict(), CHAN_WEAPON, "weapons/glauncher.wav", VOL_NORM, ATTN_NORM, 0, PITCH_NORM ); break;
			case 2: g_SoundSystem.EmitSoundDyn( pBoss.edict(), CHAN_WEAPON, "weapons/glauncher2.wav", VOL_NORM, ATTN_NORM, 0, PITCH_NORM ); break;
		}
		
		// repeat this many times
		int iMaxSpawns = 6;
		
		iSpawns++;
		pBoss.GetUserData( "grenades" ) = iSpawns;
		
		if ( iSpawns < iMaxSpawns )
			g_Scheduler.SetTimeout( "MultiGLCluster", 0.17 );
		else
		{
			pBoss.GetUserData( "grenades" ) = 0;
			g_Scheduler.SetTimeout( "HideBarrel", 1.5 );
		}
	}
}

void LinearMove( CBaseEntity@ pEntity, Vector vecDest, float flSpeed, string szCallWhenMoveDone = "", float flTriggerDelay = 0.0f )
{
	if ( g_bIntermission )
		return;
	
	// set destdelta to the vector needed to move
	Vector vecDestDelta = vecDest - pEntity.pev.origin;
	
	// divide vector length by speed to get time to reach dest
	float flTravelTime = vecDestDelta.Length() / flSpeed;
	
	// scale the destdelta vector by the time spent traveling to get velocity
	pEntity.pev.velocity = vecDestDelta / flTravelTime;
	
	g_Scheduler.SetTimeout( "LinearMoveDone", flTravelTime, EHandle( pEntity ), vecDest, szCallWhenMoveDone, flTriggerDelay );
}

void LinearMoveDone( EHandle hEntity, Vector vecDest, string szCallWhenMoveDone, float flTriggerDelay )
{
	if ( g_bIntermission )
		return;
	
	if ( hEntity.IsValid() )
	{
		CBaseEntity@ pEntity = hEntity.GetEntity();
		
		// do not.
		if ( pEntity.pev.origin.z == 450 )
			return;
		
		pEntity.pev.velocity = g_vecZero;
		g_EntityFuncs.SetOrigin( pEntity, vecDest );
		
		if ( szCallWhenMoveDone.Length() > 0 )
			g_Scheduler.SetTimeout( szCallWhenMoveDone, flTriggerDelay );
	}
}

void HideBarrel()
{
	if ( g_hBoss.IsValid() )
	{
		CBaseEntity@ pBoss = g_hBoss.GetEntity();
		
		Vector vecDest = Vector( pBoss.pev.origin.x, pBoss.pev.origin.y, 450 );
		LinearMove( pBoss, vecDest, BOSS_CRUSH_SPEED, "NextAttack", 0.5 );
	}
}

void LiftBarrel()
{
	if ( g_hBoss.IsValid() )
	{
		CBaseEntity@ pBoss = g_hBoss.GetEntity();
		
		Vector vecDest = Vector( pBoss.pev.origin.x, pBoss.pev.origin.y, 245 );
		LinearMove( pBoss, vecDest, BOSS_CRUSH_SPEED, "NextAttack", 0.5 );
	}
}

void NextAttack()
{
	if ( g_hBoss.IsValid() )
	{
		CBaseEntity@ pBoss = g_hBoss.GetEntity();
		
		@pBoss.pev.enemy = null;
		pBoss.pev.takedamage = DAMAGE_NO;
		
		switch ( g_iPhase )
		{
			case 1:
			{
				switch ( Math.RandomLong( 1, 3 ) )
				{
					case 1: CrushAttack(); break;
					case 2:
					{
						int iMinions = 0;
						CBaseEntity@ pMinion = null;
						while ( ( @pMinion = g_EntityFuncs.FindEntityByTargetname( pMinion, "giegue_minion_npc" ) ) !is null )
						{
							if ( pMinion.IsAlive() )
								iMinions++;
						}
						
						if ( iMinions == 0 )
							MinionAttack();
						else
							CrushAttack();
						break;
					}
					case 3: SingleGLAttack(); break;
				}
				break;
			}
			case 5:
			{
				int iMinions = 0;
				CBaseEntity@ pMinion = null;
				while ( ( @pMinion = g_EntityFuncs.FindEntityByTargetname( pMinion, "giegue_minion_npc" ) ) !is null )
				{
					if ( pMinion.IsAlive() )
						iMinions++;
				}
				
				if ( iMinions > 0 )
					TransformCrushAttackPre();
				else
				{
					// not hidden
					if ( pBoss.pev.origin.z != 450 )
						HideBarrel();
					else
						TransformBoidAttack( 0 );
				}
				break;
			}
			case 6:
			{
				switch ( Math.RandomLong( 1, 2 ) )
				{
					case 1: SingleGLAttack(); break;
					case 2: MultiGLAttack(); break;
				}
				break;
			}
		}
	}
}

void ExtraLifes( CBaseEntity@ pActivator, CBaseEntity@ pCaller, USE_TYPE useType, float flValue )
{
	// see if any player needs a respawn
	if ( SURVIVAL::IsEnabled() )
	{
		g_EngineFuncs.CVarSetFloat( "mp_observer_cyclic", 0.0 );
		for ( int iPlayerIndex = 1; iPlayerIndex <= g_Engine.maxClients; iPlayerIndex++ )
		{
			CBasePlayer@ pPlayer = g_PlayerFuncs.FindPlayerByIndex( iPlayerIndex );
			if ( pPlayer !is null && pPlayer.IsConnected() )
			{
				if ( !pPlayer.IsAlive() )
					PlayerRespawn( EHandle( pPlayer ) );
			}
		}
	}
	SURVIVAL::g_iTickets += TICKETS_PER_PHASE * g_PlayerFuncs.GetNumPlayers();
	SURVIVAL::ShowTickets( null );
}

void GameOver()
{
	// you are dead, dead, dead.
	g_EntityFuncs.FireTargets( "giegue_boss_music", null, null, USE_OFF, 0.0f, 0.0f );
	g_EntityFuncs.FireTargets( "giegue_gameover_bgm", null, null, USE_ON, 0.0f, 0.0f );
	
	SURVIVAL::HideTickets();
	g_EntityFuncs.FireTargets( "giegue_gameover_mm", null, null, USE_TOGGLE, 0.0f, 0.0f );
	//g_PlayerFuncs.ClientPrintAll( HUD_PRINTTALK, "* GAME OVER.\n" );
}

void IntroTextEffect( CBaseEntity@ pActivator, CBaseEntity@ pCaller, USE_TYPE useType, float flValue )
{
	CBaseEntity@ pText = g_EntityFuncs.FindEntityByTargetname( null, "giegue_intro_text" );
	string szText1 = pText.GetCustomKeyvalues().GetKeyvalue( "$s_text1" ).GetString();
	string szText2 = pText.GetCustomKeyvalues().GetKeyvalue( "$s_text2" ).GetString();
	DoTextEffect( pText, szText1, szText2, 1 );
}
void DoTextEffect( CBaseEntity@ pText, string szText1, string szText2, int iText )
{
	if ( iText == 1 )
	{
		int iChar = string( pText.pev.message ).Length();
		if ( iChar == 0 )
			g_PlayerFuncs.ClientPrintAll( HUD_PRINTCONSOLE, "HUD-MSG: \"WARNING! BOSS APPROACHING! GIANT BARREL OF DEADLY WATER\"\n" );
		
		pText.pev.message = string( pText.pev.message ) + szText1[ iChar ];
		
		if ( string( pText.pev.message ) == szText1 )
		{
			g_EntityFuncs.FireTargets( "giegue_intro_text", null, null, USE_TOGGLE, 0.0f, 0.0f );
			g_Scheduler.SetTimeout( "DoTextEffect", 2.5, @pText, szText1, szText2, 2 );
			pText.pev.message = "";
		}
		else
		{
			if ( szText1[ iChar ] != "\n" ) // keep new lines soundless
				g_EntityFuncs.FireTargets( "giegue_intro_text", null, null, USE_TOGGLE, 0.0f, 0.0f );
			
			g_Scheduler.SetTimeout( "DoTextEffect", szText1[ iChar ] == "\n" ? 0.5 : 0.05, @pText, szText1, szText2, 1 );
		}
	}
	else
	{
		int iChar = string( pText.pev.message ).Length();
		if ( iChar == 0 )
			g_PlayerFuncs.ClientPrintAll( HUD_PRINTCONSOLE, "HUD-MSG: \"ENGAGE BOILING PROTOCOLS\"\n" );
		
		pText.pev.message = string( pText.pev.message ) + szText2[ iChar ];
		g_EntityFuncs.FireTargets( "giegue_intro_text", null, null, USE_TOGGLE, 0.0f, 0.0f );
		
		if ( string( pText.pev.message ) == szText2 )
			g_EntityFuncs.FireTargets( "giegue_boss_start", null, null, USE_TOGGLE, 0.0f, 2.5f );
		else
			g_Scheduler.SetTimeout( "DoTextEffect", 0.05, @pText, szText1, szText2, 2 );
	}
}

void GameOverTXTEffect( CBaseEntity@ pActivator, CBaseEntity@ pCaller, USE_TYPE useType, float flValue )
{
	float X = 0.05;
	float Y = 0.65;
	int Z = 0;
	
	// pretend a console echo to all players
	g_PlayerFuncs.ClientPrintAll( HUD_PRINTCONSOLE, "HUD-MSG: \"GAME OVER\"\n" );
	
	g_EntityFuncs.FireTargets( "giegue_gameover_txt_l", null, null, USE_TOGGLE, 0.0f, 0.0f );
	g_EntityFuncs.FireTargets( "giegue_gameover_txt_r", null, null, USE_TOGGLE, 0.0f, 0.0f );
	
	g_Scheduler.SetTimeout( "GameOverTXTMove", 0.05, X, Y, Z );
}
void GameOverTXTMove( float X, float Y, int Z )
{
	CBaseEntity@ pTextLeft = g_EntityFuncs.FindEntityByTargetname( null, "giegue_gameover_txt_l" );
	CBaseEntity@ pTextRight = g_EntityFuncs.FindEntityByTargetname( null, "giegue_gameover_txt_r" );
	
	if ( pTextLeft is null || pTextRight is null )
		return;
	
	X += 0.05;
	if ( X > 0.45 )
	{
		X = 0.45;
		
		switch ( Z )
		{
			case 0: // down
			{
				if ( Y > 0.625 )
				{
					if ( Y > 0.65 )
						Y -= 0.010;
					else
						Y -= 0.005;
				}
				else
				{
					Y -= 0.001;
					Z = 1;
				}
				break;
			}
			case 1: // up
			{
				if ( Y < 0.675 )
				{
					if ( Y < 0.65 )
						Y += 0.010;
					else
						Y += 0.005;
				}
				else
				{
					Y += 0.001;
					Z = 0;
				}
				break;
			}
		}
	}
	
	pTextLeft.KeyValue( "x", string( X ) );
	pTextRight.KeyValue( "x", string( 0.96 - X ) );
	
	pTextLeft.KeyValue( "y", string( Y ) );
	pTextRight.KeyValue( "y", string( Y ) );
	
	g_EntityFuncs.FireTargets( "giegue_gameover_txt_l", null, null, USE_TOGGLE, 0.0f, 0.0f );
	g_EntityFuncs.FireTargets( "giegue_gameover_txt_r", null, null, USE_TOGGLE, 0.0f, 0.0f );
	
	g_Scheduler.SetTimeout( "GameOverTXTMove", X >= 0.45 ? 0.1 : 0.05, X, Y, Z );
}

// Utility code ported from AMXX
CBaseEntity@ FindEntityByOwner( CBaseEntity@ pStartEntity, string name, edict_t@ owner )
{
	CBaseEntity@ entity = pStartEntity;
	
	while ( ( @entity = g_EntityFuncs.FindEntityByString( entity, "classname", name ) ) !is null && entity.pev.owner !is owner ) { }
	
	return entity;
}

// Don't want the game to inmediatedly end as soon as tickets run out, so handle my own ticket system
namespace SURVIVAL
{

int g_iTickets = 0;
bool g_bActive = false;

const int TICKETS_HUD_CHANNEL = 1;

bool IsEnabled()
{
	return ( g_EngineFuncs.CVarGetFloat( "mp_observer_cyclic" ) == 1 );
}

void Init()
{
	g_iTickets = TICKETS_PER_PLAYER * g_PlayerFuncs.GetNumPlayers();
	g_bActive = true;
	
	ShowTickets( null );
}

void ShowTickets( CBasePlayer@ pPlayer )
{
	HUDNumDisplayParams params;
	
	params.channel = TICKETS_HUD_CHANNEL;
	params.flags = HUD_ELEM_SCR_CENTER_X | HUD_ELEM_DEFAULT_ALPHA;
	params.value = g_iTickets;
	params.x = 0;
	params.y = 1;
	params.defdigits = 3;
	params.maxdigits = 3;
	params.color1 = g_iTickets <= 1 ? RGBA_RED : RGBA_SVENCOOP;
	params.spritename = "ragemap2023/tickets.spr";
	
	g_PlayerFuncs.HudNumDisplay( pPlayer, params );
}

void HideTickets()
{
	g_PlayerFuncs.HudToggleElement( null, TICKETS_HUD_CHANNEL, false );
}

void PlayerKilled( CBasePlayer@ pPlayer, CBaseEntity@ pAttacker, int iGib )
{
	g_iTickets--;
	if ( g_iTickets < 0 )
		g_iTickets = 0;
	
	if ( g_iTickets == 0 && !IsEnabled() )
	{
		g_EntityFuncs.FireTargets( "giegue_warn_notickets", null, null, USE_ON, 0.0f, 0.0f );
		g_EngineFuncs.CVarSetFloat( "mp_observer_cyclic", 1.0 );
	}
	
	ShowTickets( null );
}

}

// Another HUD huh?
namespace TIMER
{

const int HUD_CHAN_TIMER = 2;

void Show( float flTime )
{
	HUDNumDisplayParams params;
	
	params.channel = HUD_CHAN_TIMER;
	
	params.flags = HUD_ELEM_SCR_CENTER_X | HUD_ELEM_DEFAULT_ALPHA | HUD_TIME_SECONDS | HUD_TIME_MILLISECONDS | HUD_TIME_COUNT_DOWN;
	
	params.value = flTime;
	
	params.x = 0;
	params.y = 0.06;
	
	params.color1 = RGBA_SVENCOOP;
	
	params.spritename = "stopwatch";
	
	g_PlayerFuncs.HudTimeDisplay( null, params );
	
	g_Scheduler.SetTimeout( "Hide", flTime + 1.0f );
}
void Hide()
{
	g_PlayerFuncs.HudToggleElement( null, HUD_CHAN_TIMER, false );
}

}

// Bouncing dart - Entity
const int BOLT_SPEED = 1000;
const int BOLT_NUM_BOUNCES = 18;
const int BOLT_DAMAGE = 36;

class CBounceDart : ScriptBaseEntity
{
	int m_iNumBounces;
	
	void Spawn()
	{
		//Precache();
		self.pev.movetype = MOVETYPE_BOUNCEMISSILE;
		self.pev.solid = SOLID_BBOX;
		
		g_EntityFuncs.SetModel( self, "models/crossbow_bolt.mdl" );
		
		g_EntityFuncs.SetOrigin( self, self.pev.origin );
		g_EntityFuncs.SetSize( self.pev, Vector( -2, -2, -2 ), Vector( 2, 2, 2 ) );
		
		self.pev.renderfx = kRenderFxGlowShell;
		self.pev.rendercolor = Vector( 250, 230, 50 );
		self.pev.renderamt = 255;
		self.pev.scale = 1.2;
		//self.pev.gravity = 0.01;
		
		SetTouch( TouchFunction( BoltTouch ) );
		m_iNumBounces = BOLT_NUM_BOUNCES;
	}
	
	void Precache()
	{
		g_Game.PrecacheModel( "models/crossbow_bolt.mdl" );
		g_SoundSystem.PrecacheSound( "weapons/xbow_hitbod1.wav" );
		g_SoundSystem.PrecacheSound( "weapons/xbow_hitbod2.wav" );
		g_SoundSystem.PrecacheSound( "weapons/xbow_hit1.wav" );
	}
	
	void BoltTouch( CBaseEntity@ pOther )
	{
		if ( pOther.pev.takedamage > DAMAGE_NO )
		{
			TraceResult tr = g_Utility.GetGlobalTrace();
			
			g_WeaponFuncs.ClearMultiDamage();
			pOther.TraceAttack( self.pev.owner.vars, BOLT_DAMAGE, self.pev.velocity.Normalize(), tr, DMG_DROWN );
			g_WeaponFuncs.ApplyMultiDamage( self.pev, self.pev.owner.vars );
			
			// play body "thwack" sound
			switch( Math.RandomLong( 0, 1 ) )
			{
				case 0: g_SoundSystem.EmitSound( self.edict(), CHAN_BODY, "weapons/xbow_hitbod1.wav", 1, ATTN_NORM ); break;
				case 1: g_SoundSystem.EmitSound( self.edict(), CHAN_BODY, "weapons/xbow_hitbod2.wav", 1, ATTN_NORM ); break;
			}
			
			// Hit part of the world, bounce
			if ( pOther.IsBSPModel() )
			{
				if ( m_iNumBounces > 0 )
				{
					g_EngineFuncs.VecToAngles( -( self.pev.velocity ), self.pev.angles );
					m_iNumBounces--;
				}
				else
				{
					// Unless there's is no more room for bouncing
					SetTouch( null );
					g_EntityFuncs.Remove( self );
				}
			}
			else
			{
				// Touched this guy, get out of the world
				SetTouch( null );
				g_EntityFuncs.Remove( self );
			}
		}
		else
		{
			g_SoundSystem.EmitSoundDyn( self.edict(), CHAN_BODY, "weapons/xbow_hit1.wav", Math.RandomFloat( 0.95, 1.00 ), ATTN_NORM, 0, 98 + Math.RandomLong( 0, 7 ) );
			
			if ( g_EngineFuncs.PointContents( self.pev.origin ) != CONTENTS_WATER )
			{
				g_Utility.Sparks( self.pev.origin );
			}
			
			if ( m_iNumBounces > 0 )
			{
				// Bounce around
				g_EngineFuncs.VecToAngles( -( self.pev.velocity ), self.pev.angles );
				m_iNumBounces--;
			}
			else
			{
				// That'll be enough, remove
				SetTouch( null );
				g_EntityFuncs.Remove( self );
			}
		}
	}
}

// Spawning a "bmortar" entity as-is does not properly initialize the entity
// Copy from bigmomma.cpp and pretend it's a custom entity
namespace CBMortar
{
	float m_maxFrame;
	
	void Shoot( edict_t@ pOwner, Vector vecOrigin, Vector vecVelocity )
	{
		CBaseEntity@ pSpit = g_EntityFuncs.Create( "bmortar", vecOrigin, g_vecZero, false, pOwner );
		pSpit.pev.velocity = vecVelocity;
		pSpit.pev.scale = 2.5;
		pSpit.pev.rendercolor = Vector( 250, 230, 50 );
		
		if ( m_maxFrame == 0 )
			m_maxFrame = float( g_EngineFuncs.ModelFrames( pSpit.pev.modelindex ) - 1 );
		
		g_Scheduler.SetTimeout( "Animate", 0.1, EHandle( pSpit ) );
	}
	
	void Animate( EHandle hSpit )
	{
		if ( hSpit.IsValid() )
		{
			CBaseEntity@ pSpit = hSpit.GetEntity();
			
			pSpit.pev.frame++;
			if ( pSpit.pev.frame > m_maxFrame )
				pSpit.pev.frame = 0;
			
			g_Scheduler.SetTimeout( "Animate", 0.1, EHandle( pSpit ) );
		}
	}
}

}
