namespace FLY {
    // Model and sound file paths for the fly entity
    const string MODEL_FLY = "models/ragemap2024/grunt/bee.mdl";
    const string FLY_SOUND_BUZZ1 = "ragemap2024/grunt/beesnd1.ogg";
    const string FLY_SOUND_BUZZ2 = "ragemap2024/grunt/beesnd2.ogg";
    const string FLY_SOUND_BUZZ3 = "ragemap2024/grunt/beesnd3.ogg";
    const string FLY_SOUND_DEATH = "hornet/ag_hornethit1.wav";
    const string FLY_SOUND_PAIN = "hornet/ag_hornethit2.wav";
    const string FLY_SOUND_IDLE = "hornet/ag_buzz3.wav";
    const string FLY_SOUND_ALERT = "hornet/ag_buzz2.wav";
    const string FLY_SOUND_BITE = "hornet/ag_hornethit2.wav";

    // Arrays for buzz sounds and durations
    array<string> BUZZ_SOUNDS = {FLY_SOUND_BUZZ1, FLY_SOUND_BUZZ2, FLY_SOUND_BUZZ3};
    array<string> SOUNDS_FLY = {FLY_SOUND_DEATH, FLY_SOUND_PAIN, FLY_SOUND_IDLE, FLY_SOUND_ALERT};
    array<float> BUZZ_DURATIONS = {6.25, 10.95, 5.75};

    // Constants for damage, health, and fade-out rate
    const int BUZZ_DAMAGE_FLY = 4; // Amount of damage to deal
    const int MAX_HEALTH_FLY = 60; // Maximum health
    const int FADE_OUT_RATE_FLY = 1; // Number of seconds to fade out
    // const float POISON_DAMAGE = 1.0; // Amount of damage per tick
    // const float POISON_DURATION = 3.0; // Duration of poison in seconds
    // const float POISON_INTERVAL = 1.0; // Interval between damage ticks

    void MapInit() {
        // g_Module.ScriptInfo.SetAuthor("Grunt");
        g_CustomEntityFuncs.RegisterCustomEntity("FLY::monster_fly", "monster_fly");
        g_Game.PrecacheModel(FLY::MODEL_FLY);

        for (uint i = 0; i < FLY::BUZZ_SOUNDS.length(); i++) {
            g_SoundSystem.PrecacheSound(FLY::BUZZ_SOUNDS[i]);
            g_Game.PrecacheGeneric("sound/" + FLY::BUZZ_SOUNDS[i]);
        }

        for (uint j = 0; j < FLY::SOUNDS_FLY.length(); j++) {
            g_SoundSystem.PrecacheSound(FLY::SOUNDS_FLY[j]);
            g_Game.PrecacheGeneric("sound/" + FLY::SOUNDS_FLY[j]);
        }
    }

    class monster_fly : ScriptBaseMonsterEntity {

        private float m_flNextAttack;  // Time for the next attack
        private bool m_bIsFadingOut;   // Is the fly fading out
        private float m_flSoundEndTime = 0.0;  // End time for the buzz sound
        private bool m_bPlayBuzzSound = true;  // Should the buzz sound play
        private float m_flLifespanEnd = 0.0;   // End time for the fly's lifespan
        private float m_flScale;  // Store random scale

        // CScheduledFunction@ m_pPoisonTimer = null;

        void Spawn() {
            g_EntityFuncs.SetModel(self, MODEL_FLY);  // Set the model
            g_EntityFuncs.SetSize(self.pev, Vector(-8, -8, 0), Vector(8, 8, 16));  // Set size
            self.pev.solid = SOLID_SLIDEBOX;  // Set solidity
            self.pev.movetype = MOVETYPE_FLY;  // Set movement type
            self.m_bloodColor = BLOOD_COLOR_RED;  // Set blood color
            self.pev.health = MAX_HEALTH_FLY;  // Set health
            self.pev.max_health = self.pev.health;  // Set max health
            self.pev.renderamt = 255;  // Set render amount
            self.pev.skin = 1;  // Set skin
            self.pev.rendermode = kRenderTransTexture;  // Set render mode
            self.m_MonsterState = MONSTERSTATE_IDLE;  // Set initial monster state
            self.m_FormattedName = "Plague Flies";  // Set the display name
            m_bIsFadingOut = false;  // Initial fading out state
            m_flLifespanEnd = g_Engine.time + 20.0;  // Set the lifespan end time to 20 seconds
            self.MonsterInit();  // Initialize the monster
            self.pev.nextthink = g_Engine.time + 0.1;  // Set the next think time
            SetThink(ThinkFunction(this.MonsterThink));  // Set the think function
            m_flScale = Math.RandomFloat(1.0, 1.5);  // Set a random scale
            self.pev.scale = m_flScale;  // Apply the scale
            self.pev.dmg = BUZZ_DAMAGE_FLY * m_flScale;  // Adjust damage based on scale
            self.pev.health = MAX_HEALTH_FLY * m_flScale;  // Adjust health based on scale
            self.pev.max_health = self.pev.health;  // Set the new max health
            // PrintAliveFliesCount(); // Print the number of alive flies
        }

        // void PoisonEffect(CBaseEntity@ pEntity, float elapsedTime = 0) {
        //     // Check if the entity is null, dead, or the poison duration has ended
        //     if (pEntity is null || pEntity.pev.health <= 0 || elapsedTime >= POISON_DURATION)
        //         return;

        //     // Check if the target is a player
        //     if (!pEntity.IsPlayer())
        //         return;

        //     // Check if the fly entity is valid and alive
        //     if (self is null || !self.IsAlive()) {
        //         // Stop any further poison effect if the fly is dead
        //         if (m_pPoisonTimer !is null) {
        //             g_Scheduler.RemoveTimer(m_pPoisonTimer);
        //             @m_pPoisonTimer = null;
        //         }
        //         return;
        //     }

        //     // Apply poison damage
        //     if (pEntity !is null && pEntity.IsAlive()) {
        //         pEntity.TakeDamage(self.pev, self.pev, POISON_DAMAGE, DMG_CLUB);

        //         // Schedule the next tick of poison damage
        //         @m_pPoisonTimer = g_Scheduler.SetTimeout(@this, "PoisonEffect", POISON_INTERVAL, @pEntity, elapsedTime + POISON_INTERVAL);
        //     }
        // }

        void MonsterThink() {
            if (m_bIsFadingOut)
                return;  // Skip further actions if fading out

            BaseClass.Think();
            float currentTime = g_Engine.time;  // Consolidated time-dependent checks
            if (currentTime >= m_flLifespanEnd) {
                StartFadeOut();  // Start fading out if lifespan ended
                return;
            }
            HandleBuzzSound(currentTime);  // Handle the buzz sound
            ChaseTarget(); // Chase the target
            CleanupDeadFlies(); // Clean up dead flies periodically
            self.pev.nextthink = currentTime + 0.1;  // Set the next think time
        }

        // Function to handle the buzz sound
        void HandleBuzzSound(float currentTime) {
            if (m_bPlayBuzzSound) {
                int buzzIndex = Math.RandomLong(0, BUZZ_SOUNDS.length() - 1);
                g_SoundSystem.EmitSound(self.edict(), CHAN_VOICE, BUZZ_SOUNDS[buzzIndex], 1.0, ATTN_NORM);
                m_flSoundEndTime = currentTime + BUZZ_DURATIONS[buzzIndex];  // Set the end time for the current buzz sound
                m_bPlayBuzzSound = false;  // Indicate that the buzz sound is currently playing
            }
            if (currentTime >= m_flSoundEndTime) {
                m_bPlayBuzzSound = true;  // Reset the flag to pick a new sound
            }
        }

        void UpdateYaw(Vector targetDir) {
            // Calculate the desired yaw from the direction vector
            float desiredYaw = Math.VecToYaw(targetDir);
            // Smoothly adjust the yaw angle
            self.pev.angles.y = Math.ApproachAngle(desiredYaw, self.pev.angles.y, self.pev.yaw_speed * 0.1);
        }

        int Classify() {
                return self.GetClassification( CLASS_TEAM4 );
            }

        void ChaseTarget() {
            bool foundTarget = false;
            array<CBaseEntity@> entityList(100);
            Vector vecDir;  // Direction vector to target
            float closestDist = 512.0;
            float flDist;  // Distance to target
            int iMaxRange = 512;
            Vector vecDelta = Vector(iMaxRange, iMaxRange, iMaxRange);
            
            // Counts and lists clients and monsters in the specified bounding box
            int entityCount = g_EntityFuncs.EntitiesInBox(entityList, self.pev.origin - vecDelta, self.pev.origin + vecDelta, FL_CLIENT | FL_MONSTER);

            for (int i = 0; i < entityCount; i++) {
                CBaseEntity@ pTarget = entityList[i];
                
                // Check if the target is not the fly itself, is alive, and is either a player or monster
                if (pTarget !is self && pTarget.IsAlive() && (pTarget.IsPlayer() || pTarget.IsMonster())) {
                    flDist = (pTarget.pev.origin - self.pev.origin).Length();  // Calculate the distance to the target
                    
                    // Check if the target is not an ally (enemies have a different classification)
                    if (self.IRelationship(pTarget) != R_AL && pTarget.Classify() != self.Classify()) {
                        if (flDist <= closestDist) {  // If the target is the closest one found so far, update the closest distance and direction
                            closestDist = flDist;
                            if (pTarget.IsPlayer()) {
                                // Add randomness to aim at the player's face or chest :)
                                float heightOffset = Math.RandomLong(0, 1) == 0 ? 18 : 28;  // Randomly choose between 18 (face) and 28 (chest)
                                vecDir = (pTarget.pev.origin + pTarget.pev.view_ofs - Vector(0, 0, heightOffset) - self.pev.origin).Normalize();
                            } else {
                                // General targeting for monsters
                                vecDir = (pTarget.pev.origin - self.pev.origin).Normalize();
                            }
                            self.pev.velocity = vecDir * 250;  // Set fly's velocity towards the target
                            foundTarget = true;  // Mark that a target has been found
                            
                            // Smoothly update the yaw angle to face the target
                            UpdateYaw(vecDir);
                            
                            // If the target is within attack range and the fly is ready to attack, inflict damage
                            if (flDist <= 70.0 && self.m_flNextAttack < g_Engine.time) {
                                // Stop all buzz sounds before playing the attack sound
                                for (uint j = 0; j < BUZZ_SOUNDS.length(); j++) {
                                    g_SoundSystem.StopSound(self.edict(), CHAN_VOICE, BUZZ_SOUNDS[j]);
                                }
                                pTarget.TakeDamage(self.pev, self.pev, self.pev.dmg, DMG_POISON);
                                
                                // Play a random buzz sound during the attack on the voice channel
                                int buzzIndex = Math.RandomLong(0, BUZZ_SOUNDS.length() - 1);
                                g_SoundSystem.EmitSoundDyn(self.edict(), CHAN_VOICE, BUZZ_SOUNDS[buzzIndex], 1.0, ATTN_NORM);
                                
                                // Play the bite sound on the weapon channel
                                g_SoundSystem.EmitSoundDyn(self.edict(), CHAN_WEAPON, FLY_SOUND_BITE, 1.0, ATTN_NORM);
                                self.m_flNextAttack = g_Engine.time + 0.55;  // Attack rate
                                
                                // Trigger attack animation
                                if (self.m_MonsterState != MONSTERSTATE_COMBAT || self.pev.sequence != 1) {
                                    self.m_MonsterState = MONSTERSTATE_COMBAT;
                                    self.pev.sequence = 1;
                                    self.ResetSequenceInfo();  // Reset the sequence info, ensuring the entity's current sequence animation starts from the beginning
                                }
                                self.StudioFrameAdvance();  // Advance the sequence animation frame based on the current time and frame rate
                                // PoisonEffect(pTarget);
                            } else if (self.m_MonsterState != MONSTERSTATE_COMBAT || self.pev.sequence != 0) {  
                                // Maintain idle animation if not attacking
                                self.m_MonsterState = MONSTERSTATE_IDLE;  // Set state to idle
                                self.pev.sequence = 0;
                                self.ResetSequenceInfo();
                                self.StudioFrameAdvance();
                            }
                        }
                    }
                }
            }
            
            if (!foundTarget) {
                if (self.m_MonsterState != MONSTERSTATE_IDLE) {
                    self.m_MonsterState = MONSTERSTATE_IDLE;
                    self.pev.sequence = 0;
                    self.ResetSequenceInfo();
                }
                self.StudioFrameAdvance();
            }
        }

        void StartFadeOut() {
            m_bIsFadingOut = true;
            self.pev.solid = SOLID_NOT;
            self.pev.targetname = "grunt_fading_fly";
            self.pev.rendermode = kRenderTransTexture;
            self.pev.renderamt = 255;
            SetThink(ThinkFunction(this.SUB_FadeOut));
            self.pev.nextthink = g_Engine.time + 0.1;
            // PrintAliveFliesCount();

            // Stop all active buzz sounds immediately
            for (uint i = 0; i < BUZZ_SOUNDS.length(); i++) {
                g_SoundSystem.StopSound(self.edict(), CHAN_VOICE, BUZZ_SOUNDS[i]);
            }

            // Play the death sound
            g_SoundSystem.EmitSoundDyn(self.edict(), CHAN_VOICE, FLY_SOUND_DEATH, 1.0, ATTN_NORM);
            // Stop any ongoing poison effect
            // if (m_pPoisonTimer !is null) {
            //     g_Scheduler.RemoveTimer(m_pPoisonTimer);
            //     @m_pPoisonTimer = null;
            // }
        }

        // bool messageShown = false; // Ensure message only appears once
        // int lastAliveFliesCount = 0;

        void SUB_FadeOut() {
        if (self.pev.renderamt > 2000) {
                KillAllFlies();
                // messageShown = true;
                return;
            }
            if (self.pev.renderamt > 1) {
                self.pev.renderamt = Math.max(0, self.pev.renderamt - FADE_OUT_RATE_FLY);
                self.pev.nextthink = g_Engine.time + 0.01;
                // g_Game.AlertMessage(at_console, "Fading out bug: " + self.GetClassname() + ", renderamt: " + self.pev.renderamt + "\n");
            } else {
                MarkForRemoval();
            }
        }

        void KillAllFlies() {
            CBaseEntity@ pEntity = null;
            while ((@pEntity = g_EntityFuncs.FindEntityByClassname(pEntity, "monster_fly")) !is null) {
                g_EntityFuncs.Remove(pEntity); // Remove all monster_flies
            }
            g_PlayerFuncs.ClientPrintAll(HUD_PRINTTALK, "A fly tried to break the Engine, and we don't know which one. Kill 'em all!\n");
            // PrintAliveFliesCount();
        }

        void MarkForRemoval() {
            if (!m_bIsFadingOut) {
                self.pev.flags |= FL_KILLME;
                SetThink(ThinkFunction(this.ForceRemove));
                self.pev.nextthink = g_Engine.time + 0.01;
            }
        }

        void ForceRemove() {
            if ((self.pev.flags & FL_KILLME) != 0) {
                g_EntityFuncs.Remove(self);
                // g_Game.AlertMessage(at_console, "Removed bug: " + self.GetClassname() + "\n");
                // PrintAliveFliesCount();
            }
        }

        // If somehow the bug didn't get removed and managed to break the 4th wall, try using this
        void CleanupDeadFlies() {
            CBaseEntity@ pEntity = null;
            while ((@pEntity = g_EntityFuncs.FindEntityByClassname(pEntity, "monster_fly")) !is null) {
                if (string(pEntity.pev.targetname) == "grunt_fading_fly" && (pEntity.pev.renderamt == 0 || pEntity.pev.renderamt > 255)) {
                    // Kill it with fire
                    g_EntityFuncs.Remove(pEntity);
                    // g_Game.AlertMessage(at_console, "Menace purged\n");
                    // PrintAliveFliesCount();
                }
            }
        }

        // Function to handle being killed
        void Killed(entvars_t@ pevAttacker, int iGib) {
            if (!m_bIsFadingOut) {
                self.pev.movetype = MOVETYPE_TOSS;
                self.pev.solid = SOLID_NOT;
                self.pev.targetname = "grunt_fading_fly";
                StartFadeOut();
                self.pev.sequence = 2;
                self.ResetSequenceInfo();
                self.StudioFrameAdvance();
            }
            BaseClass.Killed(pevAttacker, iGib);
            // PrintAliveFliesCount();

            // // Stop any ongoing poison effect
            // if (m_pPoisonTimer !is null) {
            //     g_Scheduler.RemoveTimer(m_pPoisonTimer);
            //     @m_pPoisonTimer = null;
            // }
        }

        // // Function to print the number of live flies
        // void PrintAliveFliesCount() {
        //     int aliveFlies = 0;
        //     int numBeesPerFly = 8; // Number of bees per fly
        //     CBaseEntity@ pEntity = null;
            
        //     while ((@pEntity = g_EntityFuncs.FindEntityByClassname(pEntity, "monster_fly")) !is null) {
        //         if (pEntity.IsAlive() && string(pEntity.pev.targetname) != "grunt_fading_fly") {
        //             aliveFlies++;
        //         }
        //     }

        //     int totalBees = aliveFlies * numBeesPerFly; // Calculate the total number of bees inside the fly entities

        //     // Only print if the count has changed
        //     if (aliveFlies != lastAliveFliesCount) {
        //         g_PlayerFuncs.ClientPrintAll(HUD_PRINTTALK, "Total entities: " + aliveFlies + " (" + totalBees + " bees each)\n");
        //         lastAliveFliesCount = aliveFlies; // Update the last recorded count
        //     }
        // }

        // Function to set yaw speed
        void SetYawSpeed() {
            self.pev.yaw_speed = 120;  // Sets how quickly the fly turns to face the target along its yaw axis (left and right turning)
        }

        // Function to handle taking damage
        int TakeDamage(entvars_t@ pevInflictor, entvars_t@ pevAttacker, float flDamage, int bitsDamageType) {
            if (bitsDamageType & DMG_BULLET > 0) {
                Vector vecDir = self.pev.origin - (pevInflictor.absmin + pevInflictor.absmax) * 0.5;
                vecDir = vecDir.Normalize();  // Calculate direction vector
                float flForce = self.DamageForce(flDamage);  // Calculate force based on damage
                self.pev.velocity = self.pev.velocity + vecDir * flForce;  // Apply force to velocity
                flDamage *= 0.7;
            }
            if (pevAttacker !is null) {
                CBaseEntity@ pAttackerEntity = g_EntityFuncs.Instance(pevAttacker);
                if (pAttackerEntity !is null) {
                    if (pAttackerEntity.IsPlayer()) {
                        CBasePlayer@ pPlayer = cast<CBasePlayer@>(pAttackerEntity);
                        float scaledDamage = flDamage / 25.0;  // Adjust scaling to 1 frag per 25 damage
                        pPlayer.pev.frags += int(scaledDamage);  // Increment frags accordingly
                    }
                }
            }
            return BaseClass.TakeDamage(pevInflictor, pevAttacker, flDamage, bitsDamageType);
        }
    }
}